def multi_as():
    """Returns anm with  input and physical as house graph"""
    import autonetkit
    import networkx as nx
    from autonetkit.load.load_json import simple_to_nx
    # returns a house graph
    data = {'directed': False,
            'graph': [('node_default',
                       {'asn': 1,
                        'device_type': 'router',
                        'host': 'localhost',
                        'ibgp_l3_cluster': u'None',
                        'ibgp_role': u'None',
                        'igp': 'ospf',
                        'ospf_area': 0,
                        'platform': 'netkit'}),
                      ('address_family', 'v4'),
                      ('file_type', 'graphml'),
                      ('enable_routing', True),
                      ('edge_default', {'ospf_cost': 1, 'type': 'physical'}),
                      ('specified_int_names', False)],
            'links': [{'dst': 'r2',
                       'dst_port': '_port1',
                       'id': 'e5',
                       'src': 'r1',
                       'src_port': '_port1',
                       'type': 'physical'},
                      {'dst': 'r8',
                       'dst_port': '_port1',
                       'id': 'e10',
                       'src': 'r10',
                       'src_port': '_port1',
                       'type': 'physical'},
                      {'dst': 'r9',
                       'dst_port': '_port1',
                       'id': 'e11',
                       'src': 'r10',
                       'src_port': '_port2',
                       'type': 'physical'},
                      {'dst': 'r2',
                       'dst_port': '_port2',
                       'id': 'e4',
                       'src': 'r3',
                       'src_port': '_port1',
                       'type': 'physical'},
                      {'dst': 'r1',
                       'dst_port': '_port2',
                       'id': 'e6',
                       'src': 'r4',
                       'src_port': '_port1',
                       'type': 'physical'},
                      {'dst': 'r2',
                       'dst_port': '_port3',
                       'id': 'e13',
                       'src': 'r4',
                       'src_port': '_port2',
                       'type': 'physical'},
                      {'dst': 'r5',
                       'dst_port': '_port1',
                       'id': 'e7',
                       'src': 'r4',
                       'src_port': '_port3',
                       'type': 'physical'},
                      {'dst': 'r7',
                       'dst_port': '_port1',
                       'id': 'e2',
                       'src': 'r4',
                       'src_port': '_port4',
                       'type': 'physical'},
                      {'dst': 'r3',
                       'dst_port': '_port2',
                       'id': 'e3',
                       'src': 'r5',
                       'src_port': '_port2',
                       'type': 'physical'},
                      {'dst': 'r6',
                       'dst_port': '_port1',
                       'id': 'e8',
                       'src': 'r5',
                       'src_port': '_port3',
                       'type': 'physical'},
                      {'dst': 'r8',
                       'dst_port': '_port2',
                       'id': 'e12',
                       'src': 'r5',
                       'src_port': '_port4',
                       'type': 'physical'},
                      {'dst': 'r8',
                       'dst_port': '_port3',
                       'id': 'e9',
                       'src': 'r6',
                       'src_port': '_port2',
                       'type': 'physical'},
                      {'dst': 'r8',
                       'dst_port': '_port4',
                       'id': 'e0',
                       'src': 'r7',
                       'src_port': '_port2',
                       'type': 'physical'},
                      {'dst': 'r9',
                       'dst_port': '_port2',
                       'id': 'e1',
                       'src': 'r7',
                       'src_port': '_port3',
                       'type': 'physical'}],
            'multigraph': False,
            'nodes': [{'asn': 1,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'A',
                       'ibgp_role': u'RRC',
                       'id': 'r1',
                       'igp': 'ospf',
                       'label': 'r1',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r1 to r2', 'id': '_port1'},
                                 {'category': 'physical', 'description': 'r1 to r4', 'id': '_port2'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r1 to r2'},
                                          2: {'category': 'physical', 'description': 'r1 to r4'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 555.0,
                       'y': 285.0},
                      {'asn': 3,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'None',
                       'ibgp_role': u'Peer',
                       'id': 'r10',
                       'igp': 'ospf',
                       'label': 'r10',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r10 to r8', 'id': '_port1'},
                                 {'category': 'physical', 'description': 'r10 to r9', 'id': '_port2'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r10 to r8'},
                                          2: {'category': 'physical', 'description': 'r10 to r9'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 765.0,
                       'y': 500.52000000000004},
                      {'asn': 1,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'A',
                       'ibgp_role': u'RRC',
                       'id': 'r2',
                       'igp': 'ospf',
                       'label': 'r2',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r2 to r1', 'id': '_port1'},
                                 {'category': 'physical',
                                  'description': 'r2 to r3', 'id': '_port2'},
                                 {'category': 'physical', 'description': 'r2 to r4', 'id': '_port3'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r2 to r1'},
                                          2: {'category': 'physical', 'description': 'r2 to r3'},
                                          3: {'category': 'physical', 'description': 'r2 to r4'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 645.0,
                       'y': 255.0},
                      {'asn': 1,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'A',
                       'ibgp_role': u'RRC',
                       'id': 'r3',
                       'igp': 'ospf',
                       'label': 'r3',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r3 to r2', 'id': '_port1'},
                                 {'category': 'physical', 'description': 'r3 to r5', 'id': '_port2'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r3 to r2'},
                                          2: {'category': 'physical', 'description': 'r3 to r5'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 765.0,
                       'y': 255.0},
                      {'asn': 1,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'A',
                       'ibgp_role': u'RR',
                       'id': 'r4',
                       'igp': 'ospf',
                       'label': 'r4',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r4 to r1', 'id': '_port1'},
                                 {'category': 'physical',
                                  'description': 'r4 to r2', 'id': '_port2'},
                                 {'category': 'physical',
                                  'description': 'r4 to r5', 'id': '_port3'},
                                 {'category': 'physical', 'description': 'r4 to r7', 'id': '_port4'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r4 to r1'},
                                          2: {'category': 'physical', 'description': 'r4 to r2'},
                                          3: {'category': 'physical', 'description': 'r4 to r5'},
                                          4: {'category': 'physical', 'description': 'r4 to r7'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 645.0,
                       'y': 345.0},
                      {'asn': 1,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'A',
                       'ibgp_role': u'RR',
                       'id': 'r5',
                       'igp': 'ospf',
                       'label': 'r5',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r5 to r4', 'id': '_port1'},
                                 {'category': 'physical',
                                  'description': 'r5 to r3', 'id': '_port2'},
                                 {'category': 'physical',
                                  'description': 'r5 to r6', 'id': '_port3'},
                                 {'category': 'physical', 'description': 'r5 to r8', 'id': '_port4'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r5 to r4'},
                                          2: {'category': 'physical', 'description': 'r5 to r3'},
                                          3: {'category': 'physical', 'description': 'r5 to r6'},
                                          4: {'category': 'physical', 'description': 'r5 to r8'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 765.0,
                       'y': 345.0},
                      {'asn': 2,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'None',
                       'ibgp_role': u'None',
                       'id': 'r6',
                       'igp': 'ospf',
                       'label': 'r6',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r6 to r5', 'id': '_port1'},
                                 {'category': 'physical', 'description': 'r6 to r8', 'id': '_port2'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r6 to r5'},
                                          2: {'category': 'physical', 'description': 'r6 to r8'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 894.616,
                       'y': 345.0},
                      {'asn': 3,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'None',
                       'ibgp_role': u'Peer',
                       'id': 'r7',
                       'igp': 'ospf',
                       'label': 'r7',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r7 to r4', 'id': '_port1'},
                                 {'category': 'physical',
                                  'description': 'r7 to r8', 'id': '_port2'},
                                 {'category': 'physical', 'description': 'r7 to r9', 'id': '_port3'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r7 to r4'},
                                          2: {'category': 'physical', 'description': 'r7 to r8'},
                                          3: {'category': 'physical', 'description': 'r7 to r9'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 645.0,
                       'y': 435.0},
                      {'asn': 3,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'None',
                       'ibgp_role': u'Peer',
                       'id': 'r8',
                       'igp': 'ospf',
                       'label': 'r8',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r8 to r10', 'id': '_port1'},
                                 {'category': 'physical',
                                  'description': 'r8 to r5', 'id': '_port2'},
                                 {'category': 'physical',
                                  'description': 'r8 to r6', 'id': '_port3'},
                                 {'category': 'physical', 'description': 'r8 to r7', 'id': '_port4'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r8 to r10'},
                                          2: {'category': 'physical', 'description': 'r8 to r5'},
                                          3: {'category': 'physical', 'description': 'r8 to r6'},
                                          4: {'category': 'physical', 'description': 'r8 to r7'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 765.0,
                       'y': 435.0},
                      {'asn': 3,
                       'device_type': 'router',
                       'host': 'localhost',
                       'ibgp_l3_cluster': u'None',
                       'ibgp_role': u'Peer',
                       'id': 'r9',
                       'igp': 'ospf',
                       'label': 'r9',
                       'ospf_area': 0,
                       'platform': 'netkit',
                       'ports': [{'category': 'loopback',
                                  'description': 'loopback',
                                  'id': 'Loopback0'},
                                 {'category': 'physical',
                                  'description': 'r9 to r10', 'id': '_port1'},
                                 {'category': 'physical', 'description': 'r9 to r7', 'id': '_port2'}],
                       'raw_interfaces': {0: {'category': 'loopback', 'description': 'loopback'},
                                          1: {'category': 'physical', 'description': 'r9 to r10'},
                                          2: {'category': 'physical', 'description': 'r9 to r7'}},
                       'specified_int_names': None,
                       'syntax': 'quagga',
                       'vrf': None,
                       'x': 645.0,
                       'y': 500.52}]}
    graph = simple_to_nx(data)
    anm = autonetkit.anm.NetworkModel()
    g_in = anm.add_overlay("input")
    g_in._replace_graph(nx.Graph(graph))
    # TODO: check if should build overlays here rather than clone in?
    g_phy = anm["phy"]
    g_phy._replace_graph(graph)
    return anm


import logging
from functools import total_ordering

import autonetkit
import autonetkit.log as log
from autonetkit.anm.interface import NmPort
from autonetkit.anm.node import NmNode
from autonetkit.log import CustomAdapter

from autonetkit.anm.ank_element import AnkElement

@total_ordering
class NmEdge(AnkElement):

    """API to access link in network"""

    def __init__(self, anm, overlay_id, src_id, dst_id, ekey=0):

        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
        object.__setattr__(self, 'src_id', src_id)
        object.__setattr__(self, 'dst_id', dst_id)
        object.__setattr__(self, 'ekey', ekey)  # for multigraphs
        #logger = logging.getLogger("ANK")
        #logstring = "Interface: %s" % str(self)
        #logger = CustomAdapter(logger, {'item': logstring})
        logger = log
        object.__setattr__(self, 'log', logger)
        self.init_logging("edge")


    def __key(self):
        """Note: key doesn't include overlay_id to allow fast cross-layer comparisons"""

        # based on http://stackoverflow.com/q/2909106

        return (self.src_id, self.dst_id)

    def __hash__(self):
        """"""

        return hash(self.__key())

    def is_multigraph(self):
        return self._graph.is_multigraph()

    def is_parallel(self):
        """If there is more than one edge between the src, dst of this edge

        >>> anm = autonetkit.topos.multi_edge()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.is_parallel()
        True
        >>> edge = anm['phy'].edge("r2", "r3")
        >>> edge.is_parallel()
        False


        """
        # TODO: check this for digraph, multiidigraph
        return self._overlay().number_of_edges(self.src, self.dst) > 1

    def __eq__(self, other):
        """

        >>> anm = autonetkit.topos.house()
        >>> e1 = anm['phy'].edge("r1", "r2")
        >>> e2 = anm['phy'].edge("r1", "r2")
        >>> e1 == e2
        True

        Can also compare across layers
        >>> e2 = anm['input'].edge("r1", "r2")
        >>> e2
        (r1, r2)
        >>> e1 == e2
        True

        For multi-edge graphs can specify the key

        >>> anm = autonetkit.topos.multi_edge()
        >>> e1 = anm['phy'].edge("r1", "r2", 0)
        >>> e2 = anm['phy'].edge("r1", "r2", 1)
        >>> e1 == e2
        False

        #TODO: compare single-edge overlays to multi-edge


        """
        if self.is_multigraph():
            try:
                if other.is_multigraph():
                    return (self.src_id, self.dst_id, self.ekey) == (other.src_id,
                                                                     other.dst_id, other.ekey)
                else:
                    # multi, single
                    return (self.src_id, self.dst_id) == (other.src_id,
                                                          other.dst_id)

            except AttributeError:
                if len(other) == 2:
                    # (src, dst)
                    return (self.src_id, self.dst_id) == other
                elif len(other) == 3:
                    # (src, dst, key)
                    return (self.src_id, self.dst_id, self.ekey) == other

        try:
            # self is single, other is single or multi -> only compare (src,
            # dst)
            return (self.src_id, self.dst_id) == (other.src_id, other.dst_id)
        except AttributeError:
            # compare to strings
            return (self.src_id, self.dst_id) == other

    def __repr__(self):
        """String of node"""
        if self.is_multigraph():
            return '(%s, %s, %s)' % (self.src,
                                         self.dst, self.ekey)

        return '(%s, %s)' % (self.src, self.dst)

    def __getitem__(self, key):
        """"""

        from autonetkit.anm.graph import NmGraph
        overlay = NmGraph(self.anm, key)
        return overlay.edge(self)

    def _overlay(self):
        from autonetkit.anm import NmGraph
        return NmGraph(self.anm, self.overlay_id)

    def __lt__(self, other):
        """

        >>> anm = autonetkit.topos.house()
        >>> e1 = anm['phy'].edge("r1", "r2")
        >>> e2 = anm['phy'].edge("r1", "r3")
        >>> e1 < e2
        True
        >>> e2 < e1
        False

        """
        if self.is_multigraph() and other.is_multigraph():
            return (self.src.node_id, self.dst.node_id, self.ekey) \
                < (other.src.node_id, other.dst.node_id, other.ekey)

        return (self.src.node_id, self.dst.node_id) \
            < (other.src.node_id, other.dst.node_id)

    # Internal properties
    def __nonzero__(self):
        """Allows for checking if edge exists

        >>> anm = autonetkit.topos.house()
        >>> e1 = anm['phy'].edge("r1", "r2")
        >>> bool(e1)
        True

        For a non-existent link, will return False
        (NOTE: doesn't throw exception)
        >>> e2 = anm['phy'].edge("r1", "r5")
        >>> bool(e2)
        False



        """
        if self.is_multigraph():
            return self._graph.has_edge(self.src_id, self.dst_id, key=self.ekey)

        return self._graph.has_edge(self.src_id, self.dst_id)

    @property
    def raw_interfaces(self):
        """Direct access to the interfaces dictionary, used by ANK modules"""
        return self._ports

    @raw_interfaces.setter
    def raw_interfaces(self, value):
        self._ports = value

    @property
    def _graph(self):
        """Return graph the edge belongs to"""

        return self.anm.overlay_nx_graphs[self.overlay_id]

    @property
    def _data(self):
        """Return data the edge belongs to"""
        if self.is_multigraph():
            return self._graph[self.src_id][self.dst_id][self.ekey]

        return self._graph[self.src_id][self.dst_id]

    # Nodes

    @property
    def src(self):
        """Source node of edge

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.src
        r1

        """

        return NmNode(self.anm, self.overlay_id, self.src_id)

    @property
    def dst(self):
        """Destination node of edge

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.dst
        r2

        """

        return NmNode(self.anm, self.overlay_id, self.dst_id)

    # Interfaces

    def apply_to_interfaces(self, attribute):
        """"

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.src_int.color = edge.dst_int.color = "blue"
        >>> (edge.src_int.color, edge.dst_int.color)
        ('blue', 'blue')
        >>> edge.color = "red"
        >>> edge.apply_to_interfaces("color")
        >>> (edge.src_int.color, edge.dst_int.color)
        ('red', 'red')
        """

        val = self.__getattr__(attribute)
        self.src_int.__setattr__(attribute, val)
        self.dst_int.__setattr__(attribute, val)

    @property
    def src_int(self):
        """Interface bound to source node of edge

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.src_int
        eth0.r1

        """

        src_int_id = self._ports[self.src_id]
        return NmPort(self.anm, self.overlay_id,
                      self.src_id, src_int_id)

    @property
    def dst_int(self):
        """Interface bound to destination node of edge

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.dst_int
        eth0.r2

        """

        dst_int_id = self._ports[self.dst_id]
        return NmPort(self.anm, self.overlay_id,
                      self.dst_id, dst_int_id)

    def bind_interface(self, node, interface):
        """Bind this edge to specified index"""

        self._ports[node.id] = interface

    def interfaces(self):
        """

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> list(edge.interfaces())
        [eth0.r1, eth0.r2]

        """

        # TODO: warn if interface doesn't exist on node

        return [NmPort(self.anm, self.overlay_id,
                       node_id, interface_id) for (node_id,
                                                   interface_id) in self._ports.items()]

    #

    def dump(self):
        return str(self._graph[self.src_id][self.dst_id])

    def get(self, key):
        """For consistency, edge.get(key) is neater than getattr(edge, key)

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.color = "red"
        >>> edge.get("color")
        'red'

        """

        return self.__getattr__(key)

    def set(self, key, val):
        """For consistency, edge.set(key, value) is neater than
        setattr(edge, key, value)

        >>> anm = autonetkit.topos.house()
        >>> edge = anm['phy'].edge("r1", "r2")
        >>> edge.color = "blue"
        >>> edge.color
        'blue'
        >>> edge.set("color", "red")
        >>> edge.color
        'red'


        """

        return self.__setattr__(key, val)

    def __getattr__(self, key):
        """Returns edge property"""
        return self._data.get(key)

    def __setattr__(self, key, val):
        """Sets edge property"""

        if key == 'raw_interfaces':
            # TODO: fix workaround for
            # http://docs.python.org/2/reference/datamodel.html#customizing-attribute-access
            object.__setattr__(self, 'raw_interfaces', val)

        self._data[key] = val



try:
    import simplejson as json
except ImportError:
    import json

import logging
import string

import autonetkit.anm
import autonetkit.log as log
import autonetkit.nidb
import autonetkit.plugins
import autonetkit.plugins.ipv4
import netaddr
import networkx as nx
from networkx.readwrite import json_graph
from autonetkit.render2 import NodeRender, PlatformRender
from autonetkit.ank import shallow_copy_nx_graph


class AnkEncoder(json.JSONEncoder):

    """Handles netaddr objects by converting to string form"""
    # TODO: look at using skipkeys = True to skip non-basic type keys (can we
    # warn on this?)

    def default(self, obj):
        if isinstance(obj, set):
            return str(obj)
        if isinstance(obj, netaddr.IPAddress):
            return str(obj)
        if isinstance(obj, netaddr.IPNetwork):
            return str(obj)
        if isinstance(obj, autonetkit.nidb.node.DmNode):
            # TODO: need to unserialize nidb nodes...
            return str(obj)
        if isinstance(obj, autonetkit.anm.node.NmNode):
            # TODO: need to unserialize nidb nodes...
            return str(obj)
        if isinstance(obj, autonetkit.anm.NmEdge):
            log.warning(
                "%s is anm overlay_edge. Use attribute rather than object in compiler." % obj)
            return str(obj)
        if isinstance(obj, autonetkit.nidb.ConfigStanza):
            retval = obj.to_json()
            return retval
        if isinstance(obj, autonetkit.render2.NodeRender):
            retval = obj.to_json()
            return retval
        if isinstance(obj, autonetkit.render2.PlatformRender):
            retval = obj.to_json()
            return retval
        if isinstance(obj, autonetkit.nidb.DmInterface):
            # TODO: check this is consistent with deserialization
            return str(obj)

        if isinstance(obj, autonetkit.anm.interface.NmPort):
                # TODO: check this is consistent with deserialization
                return str(obj)
        if isinstance(obj, nx.classes.Graph):
            # TODO: remove now?
            return json_graph.node_link_data(obj)

        if isinstance(obj, logging.LoggerAdapter):
            # TODO: filter this out in the to_json methods
            return ""

        return json.JSONEncoder.default(self, obj)


def ank_json_dumps(graph, indent=4):
    data = json_graph.node_link_data(graph)
# TODO: use regex to convert IPAddress and IPNetwork back to respective
# form in decoder
    data = json.dumps(data, cls=AnkEncoder, indent=indent, sort_keys=True)
    return data


def string_to_netaddr(val):
    retval = None
    dot_count = string.count(val, ".")
    if dot_count:
        # could be an IP or network address
        if "/" in val:
            try:
                retval = netaddr.IPNetwork(val)
            except netaddr.core.AddrFormatError:
                return  # unable to convert, leave as string
        else:
            try:
                retval = netaddr.IPAddress(val)
            except netaddr.core.AddrFormatError:
                return  # unable to convert, leave as string

    return retval


def restore_anm_nidb_from_json(data):
    # This can be used to extract from the json used to send to webserver

    d = ank_json_custom_loads(data)
    anm = autonetkit.anm.AbstractNetworkModel()
    nidb = autonetkit.nidb.DeviceModel()

    for overlay_id, overlay_data in d.items():
        if overlay_id == "nidb":
            continue  # don't restore nidb graph to anm
        anm._overlays[overlay_id] = json_graph.node_link_graph(overlay_data)

    nidb._graph = json_graph.node_link_graph(d['nidb'])
    rebind_interfaces(anm)

    return anm, nidb


def ank_json_custom_loads(data):
    # data = json.loads(data) # this is needed if dicts contain anm overlays,
    # nidb, etc
    def dict_to_object(d):
        inst = d
        for key, val in d.items():
            try:
                newval = string_to_netaddr(val)
                if newval:
                    inst[key] = newval
            except AttributeError:
                pass  # not a string
# handle lists of IP addresses
            if isinstance(val, dict) and val.get("_ConfigStanza") == True:
                val = autonetkit.nidb.ConfigStanza(**val)
                inst[key] = val  # update with (possibly) updated list

            if isinstance(val, list):
                if any(isinstance(elem, basestring) for elem in val):
                    # list contains a string
                    for index, elem in enumerate(val):
                        try:
                            new_elem = string_to_netaddr(elem)
                            if new_elem:
                                val[index] = new_elem  # in-place replacement
                        except AttributeError:
                            pass  # not a string

                inst[key] = val  # update with (possibly) updated list

        return inst

    d = json.loads(data, object_hook=dict_to_object)
    return d


def rebind_interfaces(anm):
    for overlay_id in anm.overlays():
        overlay = anm[overlay_id]
        # for edge in overlay.edges():
        #unbound_ports = edge._ports
# map nodes -> node objects, values to integers (not strings)
        # interfaces = {overlay.node(key): val for key, val in unbound_ports.items()}
        # edge._ports = interfaces # store with remapped node
        for node in overlay.nodes():
            unbound_ports = node.raw_interfaces
            if len(unbound_ports):  # is list if none set
                interfaces = {int(key): val for key, val in unbound_ports.items()}
                node.raw_interfaces = interfaces

# TODO: need to also rebind_interfaces for nidb


def rebind_nidb_interfaces(nidb):
    for node in nidb.nodes():
        unbound_ports = node.raw_interfaces
        if len(unbound_ports):  # is list if none set
            interfaces = {int(key): val for key, val in unbound_ports.items()}
            node.raw_interfaces = interfaces


def ank_json_loads(data):
    d = ank_json_custom_loads(data)
    # TODO: map back edge keys for parallel links - or is this automatic?
    return json_graph.node_link_graph(d)


def jsonify_anm(anm):
    """ Returns a dictionary of json-ified overlay graphs"""
    anm_json = {}
    for overlay_id in anm.overlays():
        NmGraph = anm[overlay_id]._graph.copy()
        for n in NmGraph:
            try:
                del NmGraph.node[n]['id']
            except KeyError:
                pass
        anm_json[overlay_id] = ank_json_dumps(NmGraph)
    return json.dumps(anm_json)


def shortened_interface(name):
    """Condenses interface name. Not canonical - mainly for brevity"""
    name = name.replace("GigabitEthernet", "ge")
    name = name.replace("0/0/0/", "")
    return name


def jsonify_anm_with_graphics(anm, nidb=None):
    """ Returns a dictionary of json-ified overlay graphs, with graphics data appended to each overlay"""
    from collections import defaultdict
    import math
    anm_json = {}
    test_anm_data = {}
    graphics_graph = anm["graphics"]._graph.copy()
    phy_graph = anm["phy"]._graph  # to access ASNs

    """simple layout of deps - more advanced layout could
    export to dot and import to omnigraffle, etc
    """
    g_deps = anm['_dependencies']
    nm_graph = g_deps._graph
    # build tree
    layers = defaultdict(list)
    nodes_by_layer = {}
    if len(nm_graph) > 0:
        topo_sort = nx.topological_sort(nm_graph)
        # trim out any nodes with no sucessors

        tree_root = topo_sort[0]
        # Note: topo_sort ensures that once reach node, would have reached its predecessors
        # start at first element after root
        for node in topo_sort:
            preds = nm_graph.predecessors(node)
            if len(preds):
                pred_level = max(nm_graph.node[p].get('level') for p in preds)
            else:
                # a root node
                pred_level = -1  # this node becomes level 0
            level = pred_level + 1
            nm_graph.node[node]['level'] = level
            layers[level].append(node)

            data = nm_graph.node[node]
            data['y'] = 100 * data['level']
            data['device_type'] = "ank_internal"

        MIDPOINT = 50  # assign either side of
        for layer, nodes in layers.items():
            # TODO: since sort is stable, first sort by parent x (avoids
            # zig-zags)
            nodes = sorted(nodes, reverse=True,
                           key=lambda x: nm_graph.degree(x))
            for index, node in enumerate(nodes):
                # TODO: work out why weird offset due to the math.pow *
                #node_x = MIDPOINT  + 125*index * math.pow(-1, index)
                node_x = MIDPOINT + 125 * index
                nm_graph.node[node]['x'] = node_x
                nodes_by_layer[node] = layer

    import random
    attribute_cache = defaultdict(dict)
    # the attributes to copy
    # TODO: check behaviour for None if explicitly set
    # TODO: need to check if attribute is set in overlay..... using API
    copy_attrs = ["x", "y", "asn", "label", "device_type", "device_subtype"]
    for node, in_data in phy_graph.nodes(data=True):
        out_data = {key: in_data.get(key) for key in copy_attrs
                    if key in in_data}
        attribute_cache[node].update(out_data)

    # Update for graphics (over-rides phy)
    for node, in_data in graphics_graph.nodes(data=True):
        out_data = {key: in_data.get(key) for key in copy_attrs
                    if key in in_data}
        attribute_cache[node].update(out_data)

        # append label from function
        for node in anm['phy']:
            attribute_cache[node.id]['label'] = str(node)

    overlay_ids = sorted(anm.overlays(),
                         key=lambda x: nodes_by_layer.get(x, 0))

    for overlay_id in overlay_ids:
        try:
            #make a shallow copy
            # input_graph = anm[overlay_id]._graph
            # nm_graph = shallow_copy_nx_graph(input_graph)
            nm_graph = anm[overlay_id]._graph.copy()
        except Exception, e:
            log.warning("Unable to copy overlay %s: %s", overlay_id, e)
            continue

        if overlay_id == "_dependencies":
            # convert to undirected for visual clarify
            nm_graph = nx.Graph(nm_graph)

        for node in nm_graph:
            node_data = dict(attribute_cache.get(node, {}))
            # update with node data from this overlay
            # TODO: check is not None won't clobber specifically set in
            # overlay...
            graph_node_data = nm_graph.node[node]
            overlay_node_data = {key: graph_node_data.get(key)
                                 for key in graph_node_data}
            node_data.update(overlay_node_data)

            # check for any non-set properties
            if node_data.get("x") is None:
                new_x = random.randint(0, 800)
                node_data['x'] = new_x
                # store for other graphs to use
                log.debug("Allocated random x %s to node %s in overlay %s" %
                          (new_x, node, overlay_id))
                attribute_cache[node]['x'] = new_x
            else:
                # cache for next time, such as vswitch in l2 for l2_bc
                attribute_cache[node]['x'] = node_data['x']
            if node_data.get("y") is None:
                new_y = random.randint(0, 800)
                node_data['y'] = new_y
                # store for other graphs to use
                attribute_cache[node]['y'] = new_y
                log.debug("Allocated random y %s to node %s in overlay %s" %
                          (new_y, node, overlay_id))
            else:
                attribute_cache[node]['y'] = node_data['y']

            # TODO: may want to re-introduce graphics to store cross-layer data for virtual nodes
            # and cache device type and device subtype
            # TODO: catch for each, if node not in cache
            try:
                attribute_cache[node]['device_type'] = node_data['device_type']
            except KeyError:
                pass  # not set
            try:
                attribute_cache[node][
                    'device_subtype'] = node_data['device_subtype']
            except KeyError:
                pass  # not set

            if node_data.get("label") == node:
                # try from cache
                node_data['label'] = attribute_cache.get(node, {}).get("label")
            if node_data.get("label") is None:
                node_data['label'] = str(node)  # don't need to cache

            # store on graph
            nm_graph.node[node] = node_data

            try:
                del nm_graph.node[node]['id']
            except KeyError:
                pass

            if nidb:
                nidb_graph = nidb.raw_graph()
                if node in nidb:
                    DmNode_data = nidb_graph.node[node]
                    try:
                        # TODO: check why not all nodes have _ports initialised
                        overlay_interfaces = nm_graph.node[node]["_ports"]
                    except KeyError:
                        continue  # skip copying interface data for this node

                    for interface_id in overlay_interfaces.keys():
                        # TODO: use raw_interfaces here
                        try:
                            nidb_interface_id = DmNode_data[
                                '_ports'][interface_id]['id']
                        except KeyError:
                            # TODO: check why arrive here - something not
                            # initialised?
                            continue
                        nm_graph.node[node]['_ports'][
                            interface_id]['id'] = nidb_interface_id
                        id_brief = shortened_interface(nidb_interface_id)
                        nm_graph.node[node]['_ports'][
                            interface_id]['id_brief'] = id_brief

        anm_json[overlay_id] = ank_json_dumps(nm_graph)
        test_anm_data[overlay_id] = nm_graph

    if nidb:
        test_anm_data['nidb'] = prepare_nidb(nidb)

    result = json.dumps(
        test_anm_data, cls=AnkEncoder, indent=4, sort_keys=True)
    return result


def prepare_nidb(nidb):
    graph = nidb.raw_graph()
    for node in graph:
        if graph.node[node].get("graphics"):
            graph.node[node]['x'] = graph.node[node]['graphics']['x']
            graph.node[node]['y'] = graph.node[node]['graphics']['y']
            graph.node[node]['device_type'] = graph.node[
                node]['graphics']['device_type']
            graph.node[node]['device_subtype'] = graph.node[
                node]['graphics']['device_subtype']

        for interface_index in graph.node[node]['_ports']:
            try:
                interface_id = graph.node[node][
                    "_ports"][interface_index]['id']
            except KeyError:  # interface doesn't exist, eg for a lan segment
                interface_id = ""
            id_brief = shortened_interface(interface_id)
            graph.node[node]["_ports"][interface_index]['id_brief'] = id_brief

    x = (graph.node[n]['x'] for n in graph)
    y = (graph.node[n]['y'] for n in graph)
    x_min = min(x)
    y_min = min(y)
    for n in graph:
        graph.node[n]['x'] += - x_min
        graph.node[n]['y'] += - y_min

    return graph


def jsonify_nidb(nidb):
    graph = prepare_nidb(nidb)
    data = ank_json_dumps(graph)
    return data


def dumps(anm, nidb=None, indent=4):
    return jsonify_anm_with_graphics(anm, nidb)



from collections import defaultdict
import os
import autonetkit.log as log


# TODO: need to work out clean way to handle template searchspaces
# and to allow users to specify their own templates from disk, or own
# searchspaces?

# TODO: need to implement caching for performance

from collections import namedtuple
src_dst = namedtuple("src_dst", ['src', 'dst'])


class NodeRender(object):

    def __init__(self, files=None, folders=None):
        self._files = defaultdict(list)
        self._folders = defaultdict(list)
        # TODO: need a new dict for each pass_id to store if memory/file... or just do all to memory?
        # and then platform decides the output
        for filename in files or []:
            self.add_file(filename)
        for foldername in files or []:
            self.add_folder(foldername)

    def __iter__(self):
        """Returns each render pass as a namedtuple of structure:

        with the folder structure flattened out to a list
        to_disk is default of False (render to memory by default)
        """
        pass

    def add_template_location(self, location):
        """TODO: allow user to specify a location to search
        """
        pass

    # TODO:implement __add__ with defaults for add_file

    # TODO: take in location (by default is autonetkit templates dir)
    def add_file(self, src, dst=None, pass_id=0):
        """Adds a file to pass "pass"
        dst is the target filename, if not set the src is used
        """
        if dst is None:
            dst = src
        self._files[pass_id].append(src_dst(src, dst))

    def add_folder(self, src, dst=None, pass_id=0):
        """Pass in folder as a list, will automatically apply os.path.sep"""
        if dst is None:
            dst = src
        self._folders[pass_id].append(src_dst(src, dst))

    def to_json(self):
        # TODO: return folders then files
        return {'files': self._files, 'folders': self._folders}

    def flatten(self):
        # flattens nested dicts out
        pass

    def get_files(self, pass_id=0):
        return self._files.get(pass_id)

    def get_file_dst(self, src, pass_id=0):
        return self._files.get(pass_id).get(src)

    def get_folders(self, pass_id=0):
        return self._folders.get(pass_id)


class PlatformRender(NodeRender):

    def __init__(self):
        super(PlatformRender, self).__init__()
        self.nodes = []
        self.base_folder = ""
        self.archive = None
        self.template_data = {}

    def add_node(self, node):
        self.nodes.append(node)

    def to_json(self):
        return self.nodes

    def __iter__(self):
        return iter(self.nodes)


def setup():
    # any setup steps for renderers
    pass


def render_node(node):
    pass


def render_topology_basic(topology):
    """Simple renderer, no caching"""
    for node in topology.render2:
        print node


# TODO: Put inside class once created
# TODO: inherit from common base

class MakoRenderer(object):

    def __init__(self):
        # if archive is none then render to memory
        # Note: this could also include setting lookup directories (unlikely to
        # be required though)
        pass

    def prepare(self, template_data):
        from mako.template import Template
        return Template(template_data)

    def render(self, template, node, version_banner, date, template_data=None):
        if template_data is None:
            template_data = {}

        try:
            return template.render(
                node=node,
                version_banner=version_banner,
                date=date,
                **template_data
            )
        except KeyError, error:
            log.warning("Unable to render %s:"
                        " %s not set" % (node, error))
            from mako import exceptions
            log.debug(exceptions.text_error_template().render())
        except AttributeError, error:
            log.warning("Unable to render %s: %s " % (node, error))
            from mako import exceptions
            log.debug(exceptions.text_error_template().render())
        except NameError, error:
            log.warning("Unable to render %s: %s. "
                        "Check all variables used are defined" % (node, error))
        except TypeError, error:
            log.warning("Unable to render %s: %s." % (node, error))
            from mako import exceptions
            log.debug(exceptions.text_error_template().render())

        import ipdb
        ipdb.set_trace()


def get_folder_contents(folder):
    retval = []

    for base, _, filenames in os.walk(folder):
        for name in filenames:
            # TODO: take in a skiplist
            if name == ".DS_Store":
                continue
            # TODO: see if can iterate folder?
            file_base = base.replace(folder, "")
            file_base = file_base[1:]  # strip leading /
            retval.append(os.path.join(file_base, name))

    return retval

extension_renderers = {'.mako': MakoRenderer()}


def render_nodes(nodes, base_folder="",
                 archive=None):

    if archive is None:
        to_memory = True
    else:
        to_memory = False

    # extracts common paths from the nodes
    # first build up list of all paths
    common_files = defaultdict(list)
    common_folders = defaultdict(list)

    if isinstance(base_folder, basestring):
        pass
    else:
        base_folder = os.path.join(*base_folder)

    # store destinations
    file_dsts = defaultdict(dict)
    folder_dsts = defaultdict(dict)

    for node in nodes:
        # print node.render2.get_pass()
        # TODO: need to iterate over all passes
        for path in node.render2.get_files():
            src = tuple(path.src)
            common_files[src].append(node)
            file_dsts[node][src] = path.dst
        # TODO: need to skip templates..... defined programatically in mapping
        # TODO: probably need to make a render class for this...
        for path in node.render2.get_folders():
            src = tuple(path.src)
            common_folders[src].append(node)
            folder_dsts[node][src] = path.dst

    # print common_files
    # print common_folders

    # load common files into cache
    # TODO: no cache now do file-by-file
    # TODO: do folders first so more specific precedence over less specific
    # in case of a clash (like in routing)

    # Prepare static classes for extensions

    # TODO: fix up naming
    for src_base, nodes in common_folders.items():
        src_folder = os.path.join(*src_base)  # "splat" list for os.path.join
        import pkg_resources
        src_path = pkg_resources.resource_filename(__name__, src_folder)
        # get structure of each folder
        contents = get_folder_contents(src_path)
        # TODO: treat per-file the same as for files below
        # TODO: could optimise so that if clobber, don't render previous
        # (unlikely to occur though)
        for filename in contents:
            out_filename = filename
            abs_filename = os.path.join(src_folder, filename)
            abs_filename = pkg_resources.resource_filename(
                __name__, abs_filename)
            with open(abs_filename) as fh:
                file_data = fh.read()

            template_renderer = None
            extension = os.path.splitext(filename)[1]
            if extension in extension_renderers:
                template_renderer = extension_renderers[extension]

                # and strip extension from output file
                out_filename = out_filename[:-len(extension)]

                import time
                version_banner = ("autonetkit_%s" %
                                  pkg_resources.get_distribution("autonetkit").version)
                date = time.strftime("%Y-%m-%d %H:%M", time.localtime())
                render_template = template_renderer.prepare(file_data)

            for node in nodes:
                # node_data = template_renderer.render(render_template,
                 # node, version_banner, date)
                if template_renderer:
                    node_data = template_renderer.render(render_template,
                                                         node, version_banner, date)
                else:
                    node_data = file_data

                # to memory -> store

                dst = folder_dsts[node][src_base]
                if isinstance(dst, basestring):
                    pass
                else:
                    dst = os.path.join(*dst)
                dst = os.path.join(dst, out_filename)
                dst = os.path.join(base_folder, dst)

                # else to file
                if to_memory:
                    node.set(dst, node_data)
                else:
                    archive.writestr(dst, node_data)

    for src, nodes in common_files.items():
        # TODO: write this to a separate function
        filename = os.path.join(*src)  # "splat" list for os.path.join
        import pkg_resources
        # TODO: specify the full file path (or provide wrapper for this to
        # specify the package)
        abs_filename = pkg_resources.resource_filename(__name__, filename)
        with open(abs_filename) as fh:
            file_data = fh.read()

        # TODO: Only do the following for .mako
        extension = os.path.splitext(filename)[1]
        template_renderer = extension_renderers[extension]

        import time
        version_banner = ("autonetkit_%s" %
                          pkg_resources.get_distribution("autonetkit").version)
        date = time.strftime("%Y-%m-%d %H:%M", time.localtime())
        render_template = template_renderer.prepare(file_data)

        # get extension

        for node in nodes:
            node_data = template_renderer.render(render_template,
                                                 node, version_banner, date)

            # to memory -> store
            dst = file_dsts[node][src]
            dst = os.path.join(base_folder, dst)

            # else to file
            if to_memory:
                node.set(dst, node_data)
            else:
                archive.writestr(dst, node_data)

    # TODO: to save memory could process for each node that has this file, etc?
    # then write that to the archive, and continue on
    # eg file by file rather than node by node?

    # TODO: if a file is greater than a certain size dont cache in memory

    # TODO: alternative would be to cache on the fly as seen, but this could s
    # another alternative is to cache if seen for second time?

# note: may want "platform render data" and a "platform renderer"


# TODO: also create a render_topology(topology)
# note that we don't have as much need to do caching as much less
# topologies than node

def render_topology(topology):
    """Pre-caches"""
    # TODO: also need to render the topology template eg lab.conf

    render_data = topology.render2
    base_folder = render_data.base_folder

    archive_filename = render_data.archive
    archive = None
    to_memory = True
    if archive_filename:
        to_memory = False
        import zipfile
        archive_filename = "%s.zip" % archive_filename
        # TODO: make rendered folder set in config, and check exists
        archive_path = os.path.join("rendered", archive_filename)
        # TODO: look at compression levels
        archive = zipfile.ZipFile(archive_path, mode='w')

    nodes = render_data.nodes
    render_nodes(nodes,
                 archive=archive,
                 base_folder=base_folder)

    if isinstance(base_folder, basestring):
        pass
    else:
        base_folder = os.path.join(*base_folder)

    for entry in render_data.get_files():
        src = entry.src
        dst = entry.dst
    # TODO: write this to a separate function
        filename = os.path.join(*src)  # "splat" list for os.path.join
        print filename
        import pkg_resources
        # TODO: specify the full file path (or provide wrapper for this to
        # specify the package)
        abs_filename = pkg_resources.resource_filename(__name__, filename)
        with open(abs_filename) as fh:
            file_data = fh.read()

        # TODO: Only do the following for .mako
        extension = os.path.splitext(filename)[1]
        template_renderer = extension_renderers[extension]

        import time
        version_banner = ("autonetkit_%s" %
                          pkg_resources.get_distribution("autonetkit").version)
        date = time.strftime("%Y-%m-%d %H:%M", time.localtime())
        render_template = template_renderer.prepare(file_data)

        # get extension
        # TODO: clean up placeholder for node renderer
        node = None
        template_data = {'topology': topology}

        topology_data = template_renderer.render(render_template,
                                                 node, date, version_banner, template_data)

        # to memory -> store
        dst = os.path.join(base_folder, dst)
        if isinstance(dst, basestring):
            pass
        else:
            dst = os.path.join(*dst)

        # else to file
        if to_memory:
            topology.set(dst, topology_data)
        else:
            archive.writestr(dst, topology_data)

    # render files

    if archive:
        archive.close()

    # warn if any folders that not supported yet for topologies


def render(nidb):
    log.info("Rendering v2")
    # TODO: should allow render to take list of nodes in case different namespace
    # perhaps allow render to take list of topologies/nodes to render?
    for topology in nidb.topologies():
        render_topology(topology)



import autonetkit.log as log
import autonetkit.ank as ank_utils

def build_layer1(anm):
    import autonetkit.design.layer1
    autonetkit.design.layer1.build_layer1(anm)

def build_layer2(anm):
    import autonetkit.design.layer2
    autonetkit.design.layer2.build_layer2(anm)

def build_layer3(anm):
    """ l3_connectivity graph: switch nodes aggregated and exploded"""
    g_in = anm['input']
    gl2_conn = anm['layer2_conn']
    g_l3 = anm.add_overlay("layer3")
    g_l3.add_nodes_from(gl2_conn, retain=['label'])
    g_l3.add_nodes_from(g_in.switches(), retain=['asn'])
    g_l3.add_edges_from(gl2_conn.edges())

    switches = g_l3.switches()

    ank_utils.aggregate_nodes(g_l3, switches)
    exploded_edges = ank_utils.explode_nodes(g_l3,
                                             switches)

    # also explode virtual switches
    vswitches = [n for n in g_l3.nodes()
                 if n['layer2'].device_type == "switch"
                 and n['layer2'].device_subtype == "virtual"]

    # explode each seperately?
    for edge in exploded_edges:
        edge.multipoint = True
        edge.src_int.multipoint = True
        edge.dst_int.multipoint = True


import collections

import time

import networkx as nx

from autonetkit.nidb.edge import DmEdge
from autonetkit.nidb.node import DmNode
from autonetkit.nidb.base import DmBase


class DmGraphData(object):

    """API to access overlay graph data in network"""

    def __init__(self, nidb):
        # Set using this method to bypass __setattr__
        object.__setattr__(self, 'nidb', nidb)

    def __repr__(self):
        return "DeviceModel data: %s" % self.nidb.raw_graph().graph

    def __getattr__(self, key):
        """Returns edge property"""
        return self.nidb.raw_graph().graph.get(key)

    def __setattr__(self, key, val):
        """Sets edge property"""
        self.nidb.raw_graph().graph[key] = val

# TODO: make this inherit same overlay base as overlay_graph for add nodes etc properties
# but not the degree etc


class DmLabTopology(object):

    """API to access lab topology in network"""
    # TODO: replace this with ConfigStanza

    def __init__(self, nidb, topology_id):
        # Set using this method to bypass __setattr__
        object.__setattr__(self, 'nidb', nidb)
        object.__setattr__(self, 'topology_id', topology_id)

    def __repr__(self):
        return "Lab Topology: %s" % self.topology_id

    @property
    def _topology_data(self):
        return self.nidb.raw_graph().graph['topologies'][self.topology_id]

    def dump(self):
        return str(self._topology_data)

    def __getattr__(self, key):
        """Returns topology property"""
        data = self._topology_data.get(key)
        return data

    def __setattr__(self, key, val):
        """Sets topology property"""
        self._topology_data[key] = val

    def set(self, key, val):
        """For consistency, topology.set(key, value) is neater
        than setattr(topology, key, value)"""
        return self.__setattr__(key, val)


class DeviceModel(DmBase):

    def __init__(self, network_model=None):
        super(DeviceModel, self).__init__()
        # only for connectivity, any other information stored on node
        if network_model and network_model['phy'].is_multigraph:
            self._graph = nx.MultiGraph()
        else:
            self._graph = nx.Graph()

        self._graph.graph['topologies'] = collections.defaultdict(dict)
        self._graph.graph['timestamp'] = time.strftime(
            "%Y%m%d_%H%M%S", time.localtime())

        if network_model:
            self._build_from_anm(network_model)

    def _build_from_anm(self, network_model):
        # TODO: Allow to specify which attributes to copy across
        # TODO: provide another function to copy across attributes
        # post-creation
        g_phy = network_model['phy']
        g_graphics = network_model['graphics']
        self.add_nodes_from(g_phy, retain=['label', 'host', 'platform',
                                           "syntax", 'Network', 'update', 'asn', ])

        self.add_edges_from(g_phy.edges())
        self.copy_graphics(network_model)

    def topology(self, key):
        return DmLabTopology(self, key)

    def topologies(self):
        return iter(DmLabTopology(self, key)
                    for key in self._graph.graph['topologies'].keys())

    @property
    def timestamp(self):
        return self._graph.graph['timestamp']

    def subgraph(self, nbunch, name=None):
        nbunch = (n.node_id for n in nbunch)  # only store the id in overlay
        return DmSubgraph(self._graph.subgraph(nbunch), name)

    def boundary_nodes(self, nbunch, nbunch2=None):
        nbunch = (n.node_id for n in nbunch)  # only store the id in overlay
        return iter(DmNode(self, node)
                    for node in nx.node_boundary(self._graph,
                                                 nbunch, nbunch2))

    def boundary_edges(self, nbunch, nbunch2=None):
        nbunch = (n.node_id for n in nbunch)  # only store the id in overlay
        return iter(DmEdge(self, src, dst)
                    for (src, dst) in nx.edge_boundary(self._graph,
                                                       nbunch, nbunch2))


class DmSubgraph(DmBase):

    def __init__(self, graph, name=None):
        super(DmSubgraph, self).__init__()
        # TODO: need to refer back to the source nidb
        # only for connectivity, any other information stored on node
        self._graph = graph
        self._name = name

    def __repr__(self):
        return "nidb: %s" % self._name



import collections
import logging
from autonetkit.log import CustomAdapter
import autonetkit.log as log


class DmInterface(object):

    def __init__(self, nidb, node_id, interface_id):
        object.__setattr__(self, 'nidb', nidb)
        object.__setattr__(self, 'node_id', node_id)
        object.__setattr__(self, 'interface_id', interface_id)
        #logger = logging.getLogger("ANK")
        #logstring = "Interface: %s" % str(self)
        #self.log = CustomAdapter(logger, {'item': logstring})
        logger = log
        object.__setattr__(self, 'log', logger)

    def __key(self):
        # based on http://stackoverflow.com/q/2909106
        return (self.interface_id, self.node_id)

    def __hash__(self):
        """"""
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()

    @property
    def is_bound(self):
        """Returns if this interface is bound to an edge on this layer"""
        return len(self.edges()) > 0

    def __repr__(self):
        description = self.description or self.interface_id
        return "%s.%s" % (self.node, description)

    def __nonzero__(self):
        """Allows for checking if node exists
        """
        return len(self._port) > 0  # if interface data set

    def __str__(self):
        return self.__repr__()

    @property
    def _graph(self):
        """Return graph the node belongs to"""
        return self.nidb.raw_graph()

    @property
    def _node(self):
        """Return node the node belongs to"""
        return self._graph.node[self.node_id]

    @property
    def _port(self):
        """Return graph the node belongs to"""
        return self._node["_ports"][self.interface_id]

    @property
    def is_loopback(self):
        """"""
        return self.category == "loopback" or self.phy.category == "loopback"

    @property
    def is_physical(self):
        """"""
        return self.category == "physical" or self.phy.category == "physical"

    @property
    def description(self):
        """"""
        return self._port.get("description")

    @property
    def is_loopback_zero(self):
        # by convention, loopback_zero is at id 0
        return self.interface_id == 0 and self.is_loopback

    @property
    def node(self):
        """Returns parent node of this interface"""
        from autonetkit.nidb import DmNode
        return DmNode(self.nidb, self.node_id)

    def dump(self):
        return str(self._port.items())

    def dict(self):
        """Returns shallow copy of dictionary used.
        Note not a deep copy: modifying values may have impact"""
        return dict(self._port.items())

    def __getattr__(self, key):
        """Returns interface property"""
        try:
            data = self._port.get(key)
        except KeyError:
            return

        if isinstance(data, dict):
            # TODO: use config stanza instead?
            return InterfaceDataDict(data)

        return data

    def get(self, key):
        """For consistency, node.get(key) is neater
        than getattr(interface, key)"""
        return getattr(self, key)

    def __setattr__(self, key, val):
        """Sets interface property"""
        try:
            self._port[key] = val
        except KeyError:
            self.set(key, val)

    def set(self, key, val):
        """For consistency, node.set(key, value) is neater
        than setattr(interface, key, value)"""
        return self.__setattr__(key, val)

    def edges(self):
        """Returns all edges from node that have this interface ID
        This is the convention for binding an edge to an interface"""
        # edges have raw_interfaces stored as a dict of {node_id: interface_id,
        # }
        valid_edges = [e for e in self.node.edges()
                       if self.node_id in e.raw_interfaces
                       and e.raw_interfaces[self.node_id] == self.interface_id]
        return valid_edges

    def neighbors(self):
        """Returns interfaces on nodes that are linked to this interface
        Can get nodes using [i.node for i in interface.neighbors()]
        """
        edges = self.edges()
        return [e.dst_int for e in edges]


class InterfaceDataDict(collections.MutableMapping):
    # TODO: replace with config stanza

    """A dictionary which allows access as dict.key as well as dict['key']
    Based on http://stackoverflow.com/questions/3387691
    only allows read only acess
    """

    def __repr__(self):
        return ", ".join(self.store.keys())

    def __init__(self, data):
        # Note this won't allow updates in place
        self.store = data
        #self.data = parent[index]
        # self.update(dict(*args, **kwargs)) # use the free update to set keys
# TODO: remove duplicate of self.store and parent

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[key] = value  # store locally

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key

    def __getattr__(self, key):
        return self.store.get(key)

    def __setattr__(self, key, value):
        if key == "store":
            object.__setattr__(self, 'store', value)
        else:
            self.store[key] = value  # store locally

    def dump(self):
        return self.store



import autonetkit
from autonetkit import example
from collections import defaultdict
from networkx.readwrite import json_graph
settings = autonetkit.config.settings


def nx_to_simple(graph):
    j_data = json_graph.node_link_data(graph)

    # map ports to their named list
    port_index_mapping = defaultdict(dict)
    for node in j_data['nodes']:
        node_id = node['id']
        _ports = node['_ports']
        ports = []
        for index, port in _ports.items():
            try:
                port_id = port["id"]
            except KeyError:
                # no port id specified
                if index == 0:
                    port_id = "Loopback0"
                else:
                    port_id = "_port%s" % index

            port_data = dict(port)
            port_data['id'] = port_id
            ports.append(port_data)
            # record the mapping of index to port_id
            port_index_mapping[node_id][index] = port_id

        node['ports'] = sorted(ports, key=lambda x: x['id'])
        del node['_ports']

    nodes = j_data['nodes']
    mapped_links = []
    for link in j_data['links']:
        src_node = nodes[link['source']]['id']
        dst_node = nodes[link['target']]['id']
        src_int_index = link['_ports'][src_node]
        dst_int_index = link['_ports'][dst_node]
        src_int = port_index_mapping[src_node][src_int_index]
        dst_int = port_index_mapping[dst_node][dst_int_index]
        # link['ports'] = {'src': src_node, 'src_port': src_int,
        #                     'dst': dst_node, 'dst_port': dst_int}
        link['src'] = src_node
        link['src_port'] = src_int
        link['dst'] = dst_node
        link['dst_port'] = dst_int
        del link['_ports']
        try:
            del link['raw_interfaces']
        except KeyError:
            pass
        del link['source']
        del link['target']

    j_data['nodes'] = sorted(j_data['nodes'], key=lambda x: x['id'])
    j_data['links'] = sorted(
        j_data['links'], key=lambda x: (x['src'], x['dst']))

    return j_data


def simple_to_nx(j_data):
    port_to_index_mapping = defaultdict(dict)
    for node in j_data['nodes']:
        if not "ports" in node:
            continue
        node_id = node['id']
        # first check for loopback zero
        ports = node['ports']
        _ports = {}  # output format
        try:
            lo_zero = [p for p in ports if p['id'] == "Loopback0"].pop()
        except IndexError:
            # can't pop -> no loopback zero, append
            lo_zero = {'category': 'loopback',
                       'description': "Loopback Zero"}
        else:
            ports.remove(lo_zero)
        finally:
            _ports[0] = lo_zero

        for index, port in enumerate(ports, start=1):
            _ports[index] = port
            port_to_index_mapping[node_id][port['id']] = index

        del node['ports']
        node['_ports'] = _ports

    nodes_by_id = {n['id']: i for i, n
                   in enumerate(j_data['nodes'])}

    unmapped_links = []

    if "links" in j_data:
        mapped_links = j_data['links']

        for link in mapped_links:
            src = link['src']
            dst = link['dst']
            src_pos = nodes_by_id[src]
            dst_pos = nodes_by_id[dst]
            src_port_id = port_to_index_mapping[src][link['src_port']]
            dst_port_id = port_to_index_mapping[dst][link['dst_port']]

            interfaces = {src: src_port_id,
                          dst: dst_port_id}

            unmapped_links.append({'source': src_pos,
                                   'target': dst_pos,
                                   '_ports': interfaces
                                   })

    j_data['links'] = unmapped_links
    return json_graph.node_link_graph(j_data)


def load_json(input_data, defaults = True):
    import json
    data = json.loads(input_data)
    graph = simple_to_nx(data)
    # TODO: any required pre-processing goes here

    if defaults:
        ank_graph_defaults = settings['JSON']['Graph Defaults']
        for (key, val) in ank_graph_defaults.items():
            if key not in graph.graph:
                graph.graph[key] = val

        ank_node_defaults = settings['JSON']['Node Defaults']
        node_defaults = graph.graph.get("node_default", {})
        for (key, val) in node_defaults.items():
            if val == 'False':
                node_defaults[key] = False

        for (key, val) in ank_node_defaults.items():
            if key not in node_defaults or node_defaults[key] == 'None':
                node_defaults[key] = val

        for node in graph:
            for (key, val) in node_defaults.items():
                if key not in graph.node[node]:
                    graph.node[node][key] = val

        ank_edge_defaults = settings['Graphml']['Edge Defaults']
        edge_defaults = graph.graph.get('edge_default', {})
        for (key, val) in ank_edge_defaults.items():
            if key not in edge_defaults or edge_defaults[key] == 'None':
                edge_defaults[key] = val

        for src, dst, data in graph.edges(data=True):
            for key, val in edge_defaults.items():
                if key not in data:
                    data[key] = val

        graph.graph['address_family'] = 'v4'
        graph.graph['enable_routing'] = True

        #TODO: move out of defaults boolean, and try/catch
        for node in sorted(graph):
            graph.node[node]['asn'] = int(graph.node[node]['asn'])

    # TODO: set default x/y if not set - or else json export will do it
    graph.graph['file_type'] = 'json'

    return graph



from autonetkit.anm import NetworkModel as NetworkModel
from autonetkit.nidb import DeviceModel as DeviceModel

from autonetkit.ank_messaging import update_vis
# for legacy compatability
from autonetkit.ank_messaging import update_vis as update_http
from autonetkit.load.initialize import initialize

import autonetkit.topologies as topos



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os

import autonetkit
import autonetkit.ank_json as ank_json
import autonetkit.config as config
import autonetkit.log as log
import autonetkit.render as render
from autonetkit.nidb import DeviceModel


import cProfile

# from https://zapier.com/engineering/profiling-python-boss/


def do_cprofile(func):
    def profiled_func(*args, **kwargs):
        profile = cProfile.Profile()
        try:
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()
            return result
        finally:
            # profile.print_stats()
            profile.dump_stats("profile")
    return profiled_func


def file_monitor(filename):
    """Generator based function to check if a file has changed"""

    last_timestamp = os.stat(filename).st_mtime

    while True:
        timestamp = os.stat(filename).st_mtime
        if timestamp > last_timestamp:
            last_timestamp = timestamp
            yield True
        yield False


#@do_cprofile
def manage_network(input_graph_string, timestamp, build=True,
                   visualise=True, compile=True, validate=True, render=True,
                   monitor=False, deploy=False, measure=False, diff=False,
                   archive=False, grid=None, ):
    """Build, compile, render network as appropriate"""

    # import build_network_simple as build_network

    import autonetkit.build_network as build_network

    if build:
        if input_graph_string:
            graph = build_network.load(input_graph_string)
        elif grid:
            graph = build_network.grid_2d(grid)

        # TODO: integrate the code to visualise on error (enable in config)
        anm = None
        try:
            anm = build_network.build(graph)
        except Exception, e:
            # Send the visualisation to help debugging
            try:
                # if visualise:
                    # import autonetkit
                    # autonetkit.update_vis(anm)
                    #TODO: refactor so only update if config or compile not both
                    pass
            except Exception, e:
                # problem with vis -> could be coupled with original exception -
                # raise original
                log.warning("Unable to visualise: %s" % e)
            raise  # raise the original exception
        else:
            if visualise:
                # log.info("Visualising network")
                # import autonetkit
                # autonetkit.update_vis(anm)
                pass

        if not compile and visualise:
            autonetkit.update_vis(anm)
            pass

        if validate:
            import autonetkit.ank_validate
            try:
                autonetkit.ank_validate.validate(anm)
            except Exception, e:
                log.warning('Unable to validate topologies: %s' % e)
                log.debug('Unable to validate topologies',
                          exc_info=True)

    if compile:
        if archive:
            anm.save()
        nidb = compile_network(anm)
        autonetkit.update_vis(anm, nidb)

        #autonetkit.update_vis(anm, nidb)
        log.debug('Sent ANM to web server')
        if archive:
            nidb.save()

        # render.remove_dirs(["rendered"])

        if render:
            import time
            #start = time.clock()
            autonetkit.render.render(nidb)
            # print time.clock() - start
            #import autonetkit.render2
            #start = time.clock()
            # autonetkit.render2.render(nidb)
            # print time.clock() - start

    if not (build or compile):

        # Load from last run

        import autonetkit.anm
        anm = autonetkit.anm.NetworkModel()
        anm.restore_latest()
        nidb = DeviceModel()
        nidb.restore_latest()
        #autonetkit.update_vis(anm, nidb)

    if diff:
        import autonetkit.diff
        nidb_diff = autonetkit.diff.nidb_diff()
        import json
        data = json.dumps(nidb_diff, cls=ank_json.AnkEncoder, indent=4)
        # log.info('Wrote diff to diff.json')

        # TODO: make file specified in config

        with open('diff.json', 'w') as fh:
            fh.write(data)

    if deploy:
        deploy_network(anm, nidb, input_graph_string)

    log.info('Configuration engine completed')  # TODO: finished what?


#@do_cprofile
def compile_network(anm):
    # log.info("Creating base network model")
    nidb = create_nidb(anm)
    g_phy = anm['phy']
    # log.info("Compiling to targets")

    for target_data in config.settings['Compile Targets'].values():
        host = target_data['host']
        platform = target_data['platform']
        if platform == 'netkit':
            import autonetkit.compilers.platform.netkit as pl_netkit
            platform_compiler = pl_netkit.NetkitCompiler(nidb, anm,
                                                         host)
        elif platform == 'VIRL':
            try:
                import autonetkit_cisco.compilers.platform.cisco as pl_cisco
                platform_compiler = pl_cisco.CiscoCompiler(nidb, anm,
                                                           host)
            except ImportError:
                log.debug('Unable to load VIRL platform compiler')
        elif platform == 'dynagen':
            import autonetkit.compilers.platform.dynagen as pl_dynagen
            platform_compiler = pl_dynagen.DynagenCompiler(nidb, anm,
                                                           host)
        elif platform == 'junosphere':
            import autonetkit.compilers.platform.junosphere as pl_junosphere
            platform_compiler = pl_junosphere.JunosphereCompiler(nidb,
                                                                 anm, host)

        if any(g_phy.nodes(host=host, platform=platform)):
            # log.info('Compiling configurations for %s on %s'
                     # % (platform, host))
            platform_compiler.compile()  # only compile if hosts set
        else:
            log.debug('No devices set for %s on %s' % (platform, host))

    return nidb


def create_nidb(anm):

    # todo: refactor this now with the layer2/layer2_bc graphs - what does nidb need?
    # probably just layer2, and then allow compiled to access layer2_bc if
    # need (eg netkit?)

    nidb = DeviceModel(anm)


    return nidb


def deploy_network(anm, nidb, input_graph_string=None):

    # log.info('Deploying Network')

    deploy_hosts = config.settings['Deploy Hosts']
    for (hostname, host_data) in deploy_hosts.items():
        for (platform, platform_data) in host_data.items():
            if not any(nidb.nodes(host=hostname, platform=platform)):
                log.debug('No hosts for (host, platform) (%s, %s), skipping deployment'
                          % (hostname, platform))
                continue

            if not platform_data['deploy']:
                log.debug('Not deploying to %s on %s' % (platform,
                                                         hostname))
                continue

            config_path = os.path.join('rendered', hostname, platform)

            if hostname == 'internal':
                try:
                    from autonetkit_cisco import deploy as cisco_deploy
                except ImportError:
                    pass  # development module, may not be available
                if platform == 'VIRL':
                    create_new_xml = False
                    if not input_graph_string:
                        create_new_xml = True  # no input, eg if came from grid
                    elif anm['input'].data['file_type'] == 'graphml':
                        create_new_xml = True  # input from graphml, create XML

                    if create_new_xml:
                        cisco_deploy.create_xml(anm, nidb,
                                                input_graph_string)
                    else:
                        cisco_deploy.package(nidb, config_path,
                                             input_graph_string)
                continue

            username = platform_data['username']
            key_file = platform_data['key_file']
            host = platform_data['host']

            if platform == 'netkit':
                import autonetkit.deploy.netkit as netkit_deploy
                tar_file = netkit_deploy.package(config_path, 'nklab')
                netkit_deploy.transfer(host, username, tar_file,
                                       tar_file, key_file)
                netkit_deploy.extract(
                    host,
                    username,
                    tar_file,
                    config_path,
                    timeout=60,
                    key_filename=key_file,
                    parallel_count=10,
                )
                if platform == 'VIRL':

                    # TODO: check why using nklab here

                    cisco_deploy.package(config_path, 'nklab')



def four_chain():
    """Returns anm with  input and physical as house graph"""
    import autonetkit
    anm = autonetkit.NetworkModel()

    g_in = anm.add_overlay("input")

    router_ids = ["r1", "r2", "r3", "r4"]
    g_in.add_nodes_from(router_ids)

    g_in.update(device_type = "router")
    g_in.update(asn = 1)

    positions = {'r1': (101, 250), 'r2': (100, 500), 'r3': (100, 750), 'r4': (100, 1000)}
    for node in g_in:
        node.x = positions[node][0]
        node.y = positions[node][1]
        eth0 = node.add_interface("eth0")
        eth0.speed = 100

    # node->node edges
    input_edges = [("r1", "r2"), ("r2", "r4"), ("r3", "r4")]
    input_interface_edges = [(g_in.node(src).interface(1), g_in.node(dst).interface(1)) for src, dst in input_edges]
    g_in.add_edges_from(input_interface_edges)

    g_phy = anm['phy']
    g_phy.add_nodes_from(g_in, retain=["device_type", "x", "y", "asn"])
    g_phy.add_edges_from(g_in.edges())

    return anm



"""Zmq based measurement server"""
# based on
# https://learning-0mq-with-pyzmq.readthedocs.org/en/latest/pyzmq/patterns/pushpull.html

# TODO: rewrite as callbacks rather than threads
import autonetkit.log as log

try:
    import zmq
except ImportError:
    log.warning("Unable to import zmq")
import json
import socket as python_socket
import telnetlib
from threading import Thread
import time
import sys


def streamer_device(port_in, port_out):
    from zmq.devices import ProcessDevice

    pd = ProcessDevice(zmq.QUEUE, zmq.PULL, zmq.PUSH)
    pd.bind_in('tcp://*:%s' % port_in)
    pd.bind_out('tcp://*:%s' % port_out)
    pd.setsockopt_in(zmq.IDENTITY, 'PULL')

    pd.setsockopt_out(zmq.IDENTITY, 'PUSH')
    pd.start()
# it will now be running in a background process


def forwarder_device(port_in, port_out):
    from zmq.devices import ProcessDevice

    pd = ProcessDevice(zmq.FORWARDER, zmq.SUB, zmq.PUB)
    pd.bind_in('tcp://*:%s' % port_in)
    pd.bind_out('tcp://*:%s' % port_out)
    pd.setsockopt_in(zmq.IDENTITY, 'SUB')
    pd.setsockopt_in(zmq.SUBSCRIBE, "")
    pd.setsockopt_out(zmq.IDENTITY, 'PUB')
    pd.start()
# it will now be running in a background process

CONNECTORS = {}

# TODO: inherit from base autonetkit connector abstract function


def netkit_connector(host, username, password, command, *args, **kwargs):
    # Note: user prompt and priv prompt have same password
    vtysh = kwargs.get("vtysh", False)

    print host, username, password, command, vtysh

    print "Connecting to %s" % (host)
    try:
        tn = telnetlib.Telnet(host, timeout=10)
    except Exception, error:
        print "Unable to connect to %s: %s" % (host, error)
        return

    tn.set_debuglevel(0)
    print "Connected to %s" % host

    welcome_banner = tn.read_until("login:", timeout=10)
    last_line = welcome_banner.splitlines()[-1]
    hostname = last_line.replace("login:", "").strip()

    linux_prompt = hostname + ":~#"

    print "Hostname is %s" % hostname

    # TODO: check why need the below for ascii/unicode/pzmq?

    tn.write(username + '\n')
    tn.read_until("Password:", timeout=10)
    tn.write(password + '\n')
    tn.read_until(linux_prompt, timeout=10)
    if vtysh:
        vtysh_prompt = hostname + "#"
        tn.write("vtysh" + "\n")
        tn.read_until(vtysh_prompt, timeout=10)
        tn.write("terminal length 0" + "\n")
        tn.read_until(vtysh_prompt, timeout=10)
        tn.write(command + "\n")
        result = tn.read_until(vtysh_prompt, timeout=10)
        tn.write("exit" + "\n")
        # TODO: check if need to parse result also to strip out prompt
    else:
        tn.write(command + "\n")
        result = tn.read_until(linux_prompt, timeout=10)
        result = "\n".join(result.splitlines()[1:-1])

    print "Finished for %s" % hostname

    tn.write("exit" + "\n")
    return hostname, result

CONNECTORS['netkit'] = netkit_connector
try:
    import autonetkit_cisco
    import autonetkit_cisco.measure_connectors
except ImportError:
    pass  # not installed
else:
    CONNECTORS[
        'iosv_ns'] = autonetkit_cisco.measure_connectors.iosv_ns_connector
    CONNECTORS[
        'csr1000v_ns'] = autonetkit_cisco.measure_connectors.iosv_ns_connector
    CONNECTORS[
        'ios_xrv_ns'] = autonetkit_cisco.measure_connectors.ios_xrv_ns_connector
    CONNECTORS[
        'nx_osv_ns'] = autonetkit_cisco.measure_connectors.nx_osv_ns_connector
    CONNECTORS[
        'ubuntu_ns'] = autonetkit_cisco.measure_connectors.linux_ns_connector


def do_connect(**kwargs):
    # TODO: use a function map
    connector = kwargs.get("connector")
    connector_fn = CONNECTORS[connector]  # TODO: capture if not found
    try:
        return connector_fn(**kwargs)
    except EOFError:
        print "Unable to connect with connector %s" % connector
        return ""


def worker():
    context = zmq.Context()
    # recieve work
    consumer_receiver = context.socket(zmq.PULL)
    consumer_receiver.connect("tcp://127.0.0.1:5560")
    # send work
    consumer_sender = context.socket(zmq.PUB)
    consumer_sender.connect("tcp://127.0.0.1:5561")
    while True:
        #  Wait for next request from client
        print "Waiting for message"
        work = consumer_receiver.recv_json()
        # socket.send(json.dumps("hello"))
        # continue
        print "Received request: ", work
        data = json.loads(work)
        host = data['host']  # TODO: rename this to host_ip
        # TODO: add support for host port (default 23)
        connector = data['connector']
        username = data['username']
        password = data['password']
        command = data['command']
        message_key = data['message_key']
        vtysh = data.get('vtysh', False)
        message_key = str(message_key)
        username = str(username)
        password = str(password)
        command = str(command)
        print "command is", command
        data = {k: str(v) for k, v in data.items()}
        try:
            hostname, result = do_connect(**data)
            success = True
        except Exception, e:
            import traceback
            traceback.print_exc()
            print e
            hostname = ""
            success = False
            result = str(e)
            if "No route to host" in e:
                # simpler message
                result = "No route to host"
            if "pexpect.TIMEOUT" in str(e):
                # TODO: test for timeout exception directly
                result = "Pexpect timeout"
        finally:
            try:
                hostname = str(hostname)
                result = str(result)
                send_data = dict(data)
                send_data.update({'command': work,
                                  "success": success,
                                  'hostname': hostname,
                                  'result': result})
                del send_data['username']
                del send_data['password']
                message = json.dumps(send_data)
            except Exception, e:
                print "cant dump", e
            else:
                consumer_sender.send_multipart([message_key, message])
                print "Sent to zmq"


def main():
    num_worker_threads = 5
    try:
        num_worker_threads = int(sys.argv[1])
    except IndexError:
        pass
    # NOTE: need pts/x available for worst-case of all threads at once
    for i in range(num_worker_threads):
        t = Thread(target=worker)
        t.daemon = True
        t.start()

    # start the streamer device
    # TODO: double check what these are used for
    streamer_device(5559, 5560)
    forwarder_device(5561, 5562)

    while True:
        time.sleep(1)

if __name__ == "__main__":
    main()



import autonetkit.ank as ank_utils
import autonetkit.config
import autonetkit.log as log
from autonetkit.ank import sn_preflen_to_network
from netaddr import IPAddress

SETTINGS = autonetkit.config.settings

#@call_log


def manual_ipv6_loopback_allocation(anm):
    """Applies manual IPv6 allocation"""

    import netaddr
    g_ipv6 = anm['ipv6']
    g_in = anm['input']

    for l3_device in g_ipv6.l3devices():
        try:
            l3_device.loopback = IPAddress(l3_device['input'].loopback_v6)
        except netaddr.AddrFormatError:
            log.debug("Unable to parse IP address %s on %s",
                l3_device['input'].loopback_v6, l3_device)

    # also need to form aggregated IP blocks (used for e.g. routing prefix
    # advertisement)
    try:
        loopback_subnet = g_in.data.ipv6_loopback_subnet
        loopback_prefix = g_in.data.ipv6_loopback_prefix
        loopback_block = sn_preflen_to_network(loopback_subnet,
                                               loopback_prefix)
    except Exception, e:
        log.info("Unable to parse specified ipv4 loopback subnets %s/%s")
    else:
        mismatched_nodes = [n for n in g_ipv6.l3devices()
                            if n.loopback and
                            n.loopback not in loopback_block]
        if len(mismatched_nodes):
            log.warning("IPv6 loopbacks set on nodes %s are not in global "
                        "loopback allocation block %s"
                        % (sorted(mismatched_nodes), loopback_block))

    loopback_blocks = {}
    for (asn, devices) in g_ipv6.groupby('asn').items():
        routers = [d for d in devices if d.is_router()]
        loopbacks = [r.loopback for r in routers]
        loopback_blocks[asn] = netaddr.cidr_merge(loopbacks)

    g_ipv6.data.loopback_blocks = loopback_blocks
    # formatted = {key: [str(v) for v in val] for key, val in loopback_blocks.items()}
    # log.info("Found loopback IP blocks %s", formatted)

#@call_log


def extract_ipv6_blocks(anm):

    # TODO: set all these blocks globally in config file, rather than repeated
    # in load, build_network, compile, etc

    from autonetkit.ank import sn_preflen_to_network
    from netaddr import IPNetwork
    g_in = anm['input']

    ipv6_defaults = SETTINGS["IP Addressing"]["v6"]

    try:
        infra_subnet = g_in.data.ipv6_infra_subnet
        infra_prefix = g_in.data.ipv6_infra_prefix
        infra_block = sn_preflen_to_network(infra_subnet, infra_prefix)
    except Exception, error:
        infra_block = IPNetwork(
            '%s/%s' % (ipv6_defaults["infra_subnet"],
                       ipv6_defaults["infra_prefix"]))
        if infra_subnet is None or infra_prefix is None:
            log.debug('Using default IPv6 infra_subnet %s', infra_block)
        else:
            log.warning('Unable to obtain IPv6 infra_subnet from input graph: %s, using default %s' % (
                error, infra_block))

    try:
        loopback_subnet = g_in.data.ipv6_loopback_subnet
        loopback_prefix = g_in.data.ipv6_loopback_prefix
        loopback_block = sn_preflen_to_network(loopback_subnet,
                                               loopback_prefix)
    except Exception, error:
        loopback_block = IPNetwork(
            '%s/%s' % (ipv6_defaults["loopback_subnet"],
                       ipv6_defaults["loopback_prefix"]))
        if loopback_subnet is None or loopback_prefix is None:
            log.debug('Using default IPv6 loopback_subnet %s',
                      loopback_block)
        else:
            log.warning('Unable to obtain IPv6 loopback_subnet from" input graph: %s, using default %s' % (
                error, loopback_block))

    try:
        vrf_loopback_subnet = g_in.data.ipv6_vrf_loopback_subnet
        vrf_loopback_prefix = g_in.data.ipv6_vrf_loopback_prefix
        vrf_loopback_block = sn_preflen_to_network(vrf_loopback_subnet,
                                                   vrf_loopback_prefix)
    except Exception, e:
        vrf_loopback_block = IPNetwork(
            '%s/%s' % (ipv6_defaults["vrf_loopback_subnet"], ipv6_defaults["vrf_loopback_prefix"]))
        if vrf_loopback_subnet is None or vrf_loopback_prefix is None:
            log.debug('Using default IPv6 vrf_loopback_subnet %s' %
                      vrf_loopback_block)
        else:
            log.warning('Unable to obtain IPv6 vrf_loopback_subnet from input graph: %s, using default %s' % (
                e, vrf_loopback_block))

    return (infra_block, loopback_block, vrf_loopback_block)

#@call_log


def manual_ipv6_infrastructure_allocation(anm):
    """Applies manual IPv6 allocation"""

    import netaddr
    g_ipv6 = anm['ipv6']
    g_in = anm['input']
    log.info('Using specified IPv6 infrastructure allocation')

    for node in g_ipv6.l3devices():
        for interface in node.physical_interfaces():
            if not interface['input'].is_bound:
                continue  # unbound interface
            if not interface['ipv6'].is_bound:
                continue
            ip_address = netaddr.IPAddress(interface['input'
                                                     ].ipv6_address)
            prefixlen = interface['input'].ipv6_prefixlen
            interface.ip_address = ip_address
            interface.prefixlen = prefixlen
            cidr_string = '%s/%s' % (ip_address, prefixlen)
            interface.subnet = netaddr.IPNetwork(cidr_string)

    broadcast_domains = [d for d in g_ipv6 if d.broadcast_domain]

    # TODO: allow this to work with specified ip_address/subnet as well as
    # ip_address/prefixlen

    global_infra_block = None
    try:
        # Note this is only pickling up if explictly set in g_in
        infra_subnet = g_in.data.ipv6_infra_subnet
        infra_prefix = g_in.data.ipv6_infra_prefix
        global_infra_block = sn_preflen_to_network(infra_subnet, infra_prefix)
    except Exception, e:
        log.info("Unable to parse specified ipv4 infra subnets %s/%s")

    from netaddr import IPNetwork
    mismatched_interfaces = []

    for coll_dom in broadcast_domains:
        connected_interfaces = [edge.dst_int for edge in
                                coll_dom.edges()]
        cd_subnets = [IPNetwork('%s/%s' % (i.subnet.network,
                                           i.prefixlen)) for i in connected_interfaces]

        if global_infra_block is not None:
            mismatched_interfaces += [i for i in connected_interfaces
            if i.ip_address not in global_infra_block]

        if len(cd_subnets) == 0:
            log.warning("Collision domain %s is not connected to any nodes",
                        coll_dom)
            continue

        try:
            assert len(set(cd_subnets)) == 1
        except AssertionError:
            mismatch_subnets = '; '.join('%s: %s/%s' % (i,
                                                        i.subnet.network, i.prefixlen) for i in
                                         connected_interfaces)
            log.warning('Non matching subnets from collision domain %s: %s',
                        coll_dom, mismatch_subnets)
        else:
            coll_dom.subnet = cd_subnets[0]  # take first entry

        # apply to remote interfaces

        for edge in coll_dom.edges():
            edge.dst_int.subnet = coll_dom.subnet

    # also need to form aggregated IP blocks (used for e.g. routing prefix
    # advertisement)
    # import autonetkit
    # autonetkit.update_vis(anm)
    if len(mismatched_interfaces):
        log.warning("IPv6 Infrastructure IPs %s are not in global "
                    "loopback allocation block %s"
                    % (sorted(mismatched_interfaces), global_infra_block))

    infra_blocks = {}
    for (asn, devices) in g_ipv6.groupby('asn').items():
        broadcast_domains = [d for d in devices if d.broadcast_domain]
        subnets = [cd.subnet for cd in broadcast_domains
                   if cd.subnet is not None]  # only if subnet is set
        infra_blocks[asn] = netaddr.cidr_merge(subnets)

    g_ipv6.data.infra_blocks = infra_blocks
    # formatted = {key: [str(v) for v in val] for key, val in infra_blocks.items()}
    # log.info("Found loopback IP blocks %s", formatted)


#@call_log


def build_ipv6(anm):
    """Builds IPv6 graph, using nodes and edges from IP graph"""
    import netaddr
    import autonetkit.plugins.ipv6 as ipv6

    # uses the nodes and edges from ipv4

    # TODO: do we also need to copy across the asn for broadcast domains?

    g_ipv6 = anm.add_overlay('ipv6')
    g_ip = anm['ip']
    g_in = anm['input']
    # retain if collision domain or not
    g_ipv6.add_nodes_from(g_ip, retain=['label', 'asn', 'allocate',
                                        'broadcast_domain'])
    g_ipv6.add_edges_from(g_ip.edges())

    # TODO: tidy up naming consitency of secondary_loopback_block and
    # vrf_loopback_block
    (infra_block, loopback_block, secondary_loopback_block) = \
        extract_ipv6_blocks(anm)

    if any(i for n in g_ip.nodes() for i in
           n.loopback_interfaces() if not i.is_loopback_zero):
        block_message = "IPv6 Secondary Loopbacks: %s" % secondary_loopback_block
        log.info(block_message)

    # TODO: replace this with direct allocation to interfaces in ip alloc
    # plugin
    allocated = sorted([n for n in g_ip if n['input'].loopback_v6])
    if len(allocated) == len(g_ip.l3devices()):
        # all allocated
        # TODO: need to infer subnetomanual_ipv6_loopback_allocation
        log.info("Using user-specified IPv6 loopback addresses")
        manual_ipv6_loopback_allocation(anm)
    else:
        log.info("Allocating from IPv6 loopback block: %s" % loopback_block)
        if len(allocated):
            log.warning(
                "Using automatic IPv6 loopback allocation. IPv6 loopback addresses specified on nodes %s will be ignored." % allocated)
        else:
            log.info("Automatically assigning IPv6 loopback addresses")

        ipv6.allocate_loopbacks(g_ipv6, loopback_block)

    l3_devices = [d for d in g_in if d.device_type in ('router', 'server')]

    manual_alloc_devices = set()
    for device in l3_devices:
        physical_interfaces = list(device.physical_interfaces())
        allocated = list(
            interface.ipv6_address for interface in physical_interfaces
            if interface.is_bound and interface['ipv6'].is_bound
            and interface['ip'].allocate is not False)

        #TODO: check for repeated code

#TODO: copy allocate from g_ip to g_ipv6
        if all(interface.ipv6_address for interface in
               physical_interfaces if interface.is_bound
               and interface['ipv6'].is_bound
               and interface['ip'].allocate is not False):
            # add as a manual allocated device
            manual_alloc_devices.add(device)

    if manual_alloc_devices == set(l3_devices):
        log.info("Using user-specified IPv6 infrastructure addresses")
        manual_alloc_ipv6_infrastructure = True
    else:
        log.info("Allocating from IPv6 Infrastructure block: %s" % infra_block)
        manual_alloc_ipv6_infrastructure = False
        # warn if any set
        allocated = []
        unallocated = []
        for node in l3_devices:
            allocated += sorted([i for i in node.physical_interfaces()
                                 if i.is_bound and i.ipv6_address])
            unallocated += sorted([i for i in node.physical_interfaces()
                                   if i.is_bound and not i.ipv6_address
                                   and i['ipv6'].is_bound])

        # TODO: what if IP is set but not a prefix?
        if len(allocated):
            # TODO: if set is > 50% of nodes then list those that are NOT set
            log.warning(
                "Using automatic IPv6 interface allocation. IPv6 interface addresses specified on interfaces %s will be ignored." % allocated)
        else:
            log.info("Automatically assigning IPv6 infrastructure addresses")

    if manual_alloc_ipv6_infrastructure:
        manual_ipv6_infrastructure_allocation(anm)
    else:
        ipv6.allocate_infra(g_ipv6, infra_block)
        # TODO: see if this is still needed or if can allocate direct from the
        # ipv6 allocation plugin
        for node in g_ipv6.l3devices():
            for interface in node:
                edges = list(interface.edges())
                if len(edges):
                    edge = edges[0]  # first (only) edge
                    # TODO: make this consistent
                    interface.ip_address = edge.ip
                    interface.subnet = edge.dst.subnet  # from collision domain

    ipv6.allocate_secondary_loopbacks(g_ipv6, secondary_loopback_block)
    for node in g_ipv6:
        node.static_routes = []

    for node in g_ipv6.routers():
        # TODO: test this code
        node.loopback_zero.ip_address = node.loopback
        node.loopback_zero.subnet = netaddr.IPNetwork("%s/32" % node.loopback)
        for interface in node.loopback_interfaces():
            if not interface.is_loopback_zero:
                # TODO: fix this inconsistency elsewhere
                interface.ip_address = interface.loopback

#@call_log



#!/usr/bin/python
# -*- coding: utf-8 -*-
from autonetkit.compiler import sort_sessions
from autonetkit.compilers.device.device_base import DeviceCompiler
from autonetkit.nidb import ConfigStanza

import autonetkit.plugins.naming as naming
import autonetkit.log as log
import netaddr
from autonetkit.ank import sn_preflen_to_network


class RouterCompiler(DeviceCompiler):

    """Base router compiler"""

    lo_interface = 'lo0'
    lo_interface_prefix = 'lo'

# and set per platform

    def ibgp_session_data(self, session, ip_version):

        # Don't make staticmethod as may want to extend (DeviceCompiler.g. in
        # IOS compiler for vpnv4)

        node = session.src
        neigh = session.dst

        if ip_version == 4:
            neigh_ip = neigh['ipv4']
            use_ipv4 = True
            use_ipv6 = False
        elif ip_version == 6:
            neigh_ip = neigh['ipv6']
            use_ipv4 = False
            use_ipv6 = True

        # TODO: return ConfigStanza rather than a dict
        data = {  # TODO: this is platform dependent???
            'neighbor': neigh.label,
            'use_ipv4': use_ipv4,
            'use_ipv6': use_ipv6,
            'asn': neigh.asn,
            'loopback': neigh_ip.loopback,
            'update_source': node.loopback_zero.id,
        }
        return data

    def ebgp_session_data(self, session, ip_version):
        node = session.src
        neigh = session.dst
        if ip_version == 4:
            neigh_ip = neigh['ipv4']
            local_int_ip = session.src_int['ipv4'].ip_address
            dst_int_ip = session.dst_int['ipv4'].ip_address
            use_ipv4 = True
            use_ipv6 = False
        elif ip_version == 6:
            neigh_ip = neigh['ipv6']
            local_int_ip = session.src_int['ipv6'].ip_address
            dst_int_ip = session.dst_int['ipv6'].ip_address
            use_ipv4 = False
            use_ipv6 = True

        md5_password = session.md5_password
        multihop = session.multihop

        # TODO: return ConfigStanza rather than a dict
        data = {  # TODO: change templates to access from node.bgp.lo_int
            'neighbor': neigh.label,
            'use_ipv4': use_ipv4,
            'use_ipv6': use_ipv6,
            'asn': neigh.asn,
            'loopback': neigh_ip.loopback,
            'local_int_ip': local_int_ip,
            'dst_int_ip': dst_int_ip,
            'update_source': node.loopback_zero.id,
            'md5_password': md5_password,
            'multihop': multihop,
        }
        return data

    def __init__(self, nidb, anm):
        """Base Router compiler"""

        super(RouterCompiler, self).__init__(nidb, anm)

    def compile(self, node):
        node.do_render = True  # turn on rendering

        phy_node = self.anm['phy'].node(node)
        ipv4_node = self.anm['ipv4'].node(node)

        node.global_custom_config = phy_node.custom_config

        node.add_stanza("ip")
        node.ip.use_ipv4 = phy_node.use_ipv4 or False
        node.ip.use_ipv6 = phy_node.use_ipv6 or False
        if not (node.ip.use_ipv4 and node.ip.use_ipv6):
            node.log.debug(
                'Neither IPv4 nor IPv6 specified: using default IPv4')

        self.static_routes(node)

        # node.ip.use_ipv4 = True

        node.label = naming.network_hostname(phy_node)
        node.input_label = phy_node.id
        if node.ip.use_ipv4:
            node.loopback = ipv4_node.loopback
            node.loopback_subnet = netaddr.IPNetwork(node.loopback)
            node.loopback_subnet.prefixlen = 32

        if self.anm['phy'].data.enable_routing:
            # applies even if ipv4 disabled, used for eg eigrp, bgp, ...
            node.router_id = ipv4_node.loopback

        self.interfaces(node)
        if self.anm.has_overlay('ospf') and node in self.anm['ospf']:
            self.ospf(node)
        if self.anm.has_overlay('isis') and node in self.anm['isis']:
            self.isis(node)
        if self.anm.has_overlay('eigrp') and node in self.anm['eigrp']:
            self.eigrp(node)
        ###self.rip(node)
        if self.anm.has_overlay('rip') and node in self.anm['rip']:
            self.rip(node)
        # TODO: drop bgp overlay
        bgp_overlays = ["bgp", "ebgp_v4", "ibgp_v4", "ebgp_v6", "ibgp_v6"]
        use_bgp = False
        for overlay in bgp_overlays:
            if (self.anm.has_overlay(overlay)
                        and node in self.anm[overlay]
                        and self.anm[overlay].node(node).degree() > 0
                    ):
                use_bgp = True
                break

        if use_bgp:
            self.bgp(node)

    def static_routes(self, node):
        node.ipv4_static_routes = []
        if self.anm.has_overlay("ipv4"):
            ipv4_node = self.anm['ipv4'].node(node)
            if ipv4_node and ipv4_node.static_routes:
                for route in ipv4_node.static_routes:
                    stanza = ConfigStanza(
                        prefix=route.prefix,
                        netmask=route.netmask,
                        nexthop=route.nexthop,
                        metric=route.metric,
                    )
                    node.ipv4_static_routes.append(stanza)

        node.ipv6_static_routes = []
        if self.anm.has_overlay("ipv6"):
            ipv6_node = self.anm['ipv6'].node(node)
            if ipv6_node and ipv6_node.static_routes:
                for route in ipv6_node.static_routes:
                    stanza = ConfigStanza(
                        prefix=route.prefix,
                        nexthop=route.nexthop,
                        metric=route.metric,
                    )
                    node.ipv6_static_routes.append(stanza)

    def interfaces(self, node):
        node.interfaces = []

        node.loopback_zero.id = self.lo_interface
        node.loopback_zero.description = 'Loopback'
        phy_node = self.anm['phy'].node(node)
        node.loopback_zero.custom_config = phy_node.loopback_zero.custom_config

        if node.ip.use_ipv4:
            ipv4_node = self.anm['ipv4'].node(node)
            node.loopback_zero.ipv4_address = ipv4_node.loopback
            node.loopback_zero.ipv4_subnet = node.loopback_subnet

        # TODO: bne consistent wit hcidr name so can use in cisco ios xr templates
        # if node.ip.use_ipv6:
            #ipv6_node = self.anm['ipv6'].node(node)
            #node.loopback_zero.ipv6_address = ipv6_node.loopback
            #node.loopback_zero.ipv6_subnet = node.loopback_subnet

        for interface in node.physical_interfaces():
            phy_int = self.anm['phy'].interface(interface)
            interface.physical = True

            # TODO: allocate ID in platform compiler

            if not phy_int:
                # for instance if added as management interface to nidb in
                # compile
                continue

            interface.custom_config = phy_int.custom_config

            interface.description = phy_int.description

            continue

    def ospf(self, node):
        """Returns OSPF links, also sets process_id
        """

        g_ospf = self.anm['ospf']
        g_ipv4 = self.anm['ipv4']
        ospf_node = g_ospf.node(node)
        ospf_stanza = node.add_stanza("ospf")

        ospf_stanza.custom_config = ospf_node.custom_config

        # default, inherited enable if necessary
        node.ospf.ipv4_mpls_te = False

        node.ospf.loopback_area = g_ospf.node(node).area or 0

        node.ospf.process_id = ospf_node.process_id
        node.ospf.lo_interface = self.lo_interface

        node.ospf.ospf_links = []

        # aggregate by area
        from collections import defaultdict
        interfaces_by_area = defaultdict(list)

        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # don't configure IGP for this interface

            ospf_int = g_ospf.interface(interface)
            if ospf_int and ospf_int.is_bound:
                area = ospf_int.area
                # TODO: can we remove the next line?
                area = str(area)  # can't serialize IPAddress object to JSON
                # TODO: put in interface rather than interface.id for
                # consistency
                try:
                    cost = int(ospf_int.cost)
                except TypeError:
                    #TODO: log warning here
                    cost = 1
                stanza = ConfigStanza(id=interface.id,
                                      cost=cost, passive=False)

                if node.ip.use_ipv4:
                    stanza.ipv4_address = ospf_int['ipv4'].ip_address
                    stanza.ipv4_subnet = ospf_int['ipv4'].subnet
                if node.ip.use_ipv6:
                    stanza.ipv6_address = ospf_int['ipv6'].ip_address
                    stanza.ipv6_subnet = ospf_int['ipv6'].subnet

                interfaces_by_area[area].append(stanza)

        loopback_zero = node.loopback_zero
        ospf_loopback_zero = g_ospf.interface(loopback_zero)
        router_area = ospf_loopback_zero.area  # area assigned to router
        # can't serialize IPAddress object to JSON
        router_area = str(router_area)
        stanza = ConfigStanza(id=node.loopback_zero.id,
                              cost=0, passive=True)
        interfaces_by_area[router_area].append(stanza)

        node.ospf.interfaces_by_area = ConfigStanza(**interfaces_by_area)

        # TODO: split this into a generic IGP function
        added_networks = set()
        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # don't configure IGP for this interface
            if not interface.use_ipv4:
                continue
            ipv4_int = g_ipv4.interface(interface)
            ospf_int = g_ospf.interface(interface)
            if not ospf_int.is_bound:
                continue  # not an OSPF interface
            try:
                ospf_cost = int(ospf_int.cost)
            except TypeError:
                try:
                    ospf_cost = netaddr.IPAddress(ospf_int.cost)
                except (TypeError, netaddr.AddrFormatError):
                    log.debug('Using default OSPF cost of 1 for %s on %s',
                              ospf_int, node)
                    ospf_cost = 1  # default
            interface.ospf_cost = ospf_cost
            network = ipv4_int.subnet

            #TODO: refactor to allow injecting loopback IPs, etc into IGP
            if ospf_int and ospf_int.is_bound and network \
                    not in added_networks:  # don't add more than once
                added_networks.add(network)
                link_stanza = ConfigStanza(
                    network=network, interface=interface, area=ospf_int.area)
                node.ospf.ospf_links.append(link_stanza)

        for interface in node.loopback_interfaces():
            phy_int = self.anm['phy'].interface(interface)
            ipv4_int = g_ipv4.interface(interface)
            if not phy_int.inject_to_igp:
                #TODO: need to confirm which area to use
                continue

            network = ipv4_int.subnet
            if network in added_networks:
                #TODO: may want to warn here
                continue # already advertised ("how? - warn if so!)

            # Use the same area as Loopback Zero
            area = node.ospf.loopback_area

            if not network:
                log.info("Not injecting unset network on loopback %s "
                    "to IGP", interface)
                continue

            link_stanza = ConfigStanza(
             network=network, interface=interface, area=area)
            node.ospf.ospf_links.append(link_stanza)

            # also add networks for subnets to servers in the same AS

    def bgp(self, node):
        phy_node = self.anm['phy'].node(node)
        g_ipv4 = self.anm['ipv4']
        if node.ip.use_ipv6:
            g_ipv6 = self.anm['ipv6']
        asn = phy_node.asn
        node.asn = asn
        bgp_stanza = node.add_stanza("bgp")
        if self.anm.has_overlay("bgp"):
            bgp_stanza.custom_config = phy_node['bgp'].custom_config
        else:
            bgp_stanza.custom_config = ""

        node.bgp.ipv4_advertise_subnets = []
        if node.ip.use_ipv4:
            node.bgp.ipv4_advertise_subnets = \
                g_ipv4.data.infra_blocks.get(
                    asn) or []  # could be none (one-node AS) default empty list
        node.bgp.ipv6_advertise_subnets = []
        if node.ip.use_ipv6:
            g_ipv6 = self.anm['ipv6']
            node.bgp.ipv6_advertise_subnets = \
                g_ipv6.data.infra_blocks.get(asn) or []

        ibgp_neighbors = []
        ibgp_rr_clients = []
        ibgp_rr_parents = []

        g_ibgp_v4 = self.anm['ibgp_v4']
        for session in sort_sessions(g_ibgp_v4.edges(phy_node)):
            if session.exclude:
                log.debug('Skipping excluded ibgp session %s', session)
                # exclude from regular ibgp config (eg VRF, VPLS, etc)
                continue

            data = self.ibgp_session_data(session, ip_version=4)
            bgp_stanza = ConfigStanza(**data)

            direction = session.direction
            if direction == 'down':
                ibgp_rr_clients.append(bgp_stanza)
            elif direction == 'up':
                ibgp_rr_parents.append(bgp_stanza)
            else:
                ibgp_neighbors.append(bgp_stanza)

        # TODO: check v6 hierarchy only created if node set to being v4 or v6

        if node.ip.use_ipv6:
            g_ibgp_v6 = self.anm['ibgp_v6']
            for session in sort_sessions(g_ibgp_v6.edges(phy_node)):
                if session.exclude:
                    log.debug('Skipping excluded ibgp session %s', session)
                    # exclude from regular ibgp config (eg VRF, VPLS, etc)
                    continue
                data = self.ibgp_session_data(session, ip_version=6)
                bgp_stanza = ConfigStanza(**data)

                direction = session.direction
                if direction == 'down':
                    ibgp_rr_clients.append(bgp_stanza)
                elif direction == 'up':
                    ibgp_rr_parents.append(bgp_stanza)
                else:
                    ibgp_neighbors.append(bgp_stanza)

        # TODO: update this to use ibgp_v4 and ibgp_v6 overlays

        node.bgp.ibgp_neighbors = ibgp_neighbors
        node.bgp.ibgp_rr_clients = ibgp_rr_clients
        node.bgp.ibgp_rr_parents = ibgp_rr_parents

        # ebgp

        ebgp_neighbors = []
        g_ebgp_v4 = self.anm['ebgp_v4']
        for session in sort_sessions(g_ebgp_v4.edges(phy_node)):
            if session.exclude:
                log.debug('Skipping excluded ebgp session %s', session)
                # exclude from regular ibgp config (eg VRF, VPLS, etc)
                continue
            data = self.ebgp_session_data(session, ip_version=4)
            bgp_stanza = ConfigStanza(**data)
            ebgp_neighbors.append(bgp_stanza)

        if node.ip.use_ipv6:
            g_ebgp_v6 = self.anm['ebgp_v6']
            for session in sort_sessions(g_ebgp_v6.edges(phy_node)):
                if session.exclude:
                    log.debug('Skipping excluded ebgp session %s', session)
                    # exclude from regular ibgp config (eg VRF, VPLS, etc)
                    continue
                data = self.ebgp_session_data(session, ip_version=6)
                bgp_stanza = ConfigStanza(**data)
                ebgp_neighbors.append(bgp_stanza)

        ebgp_neighbors = sorted(ebgp_neighbors, key=lambda x: x.asn)
        node.bgp.ebgp_neighbors = ebgp_neighbors

        return

    def rip(self, node):
        g_rip = self.anm['rip']
        g_ipv4 = self.anm['ipv4']
        rip_node = self.anm['rip'].node(node)
        node.rip.process_id = rip_node.process_id
        node.rip.custom_config = rip_node.custom_config

        ipv4_networks = set()
        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # discontinue configuring IGP for this interface. 
            ipv4_int = g_ipv4.interface(interface)
            rip_int = g_rip.interface(interface)
            if not rip_int.is_bound:
                continue  # not an rip interface
            network = ipv4_int.subnet
            if rip_int and rip_int.is_bound and interface.use_ipv4:
                ipv4_networks.add(network)

        ipv4_networks.add(node.loopback_zero.ipv4_cidr)
        node.rip.ipv4_networks = sorted(list(ipv4_networks))


    def isis(self, node):
        g_isis = self.anm['isis']
        isis_node = g_isis.node(node)
        process_id = isis_node.process_id
        node.isis.custom_config = isis_node.custom_config

        # default, inherited enable if necessary
        node.isis.ipv4_mpls_te = False

        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # don't configure IGP for this interface

            phy_int = self.anm['phy'].interface(interface)

            isis_int = phy_int['isis']
            if isis_int and isis_int.is_bound:
                isis_node = g_isis.node(node)
                interface.isis = {
                    'metric': isis_int.metric,
                    'process_id': process_id,
                    'use_ipv4': node.ip.use_ipv4,
                    'use_ipv6': node.ip.use_ipv6,
                    'multipoint': isis_int.multipoint,
                }

                # TODO: add wrapper for this

        g_isis = self.anm['isis']
        isis_node = self.anm['isis'].node(node)
        node.isis.net = isis_node.net

        # TODO: generalise loopbacks to allow more than one per device

        node.isis.process_id = process_id
        node.isis.lo_interface = self.lo_interface

# set isis on loopback_zero

        node.loopback_zero.isis = {'use_ipv4': node.ip.use_ipv4,
                                   'use_ipv6': node.ip.use_ipv6}

        # TODO: add wrapper for this

    def eigrp(self, node):

        g_eigrp = self.anm['eigrp']
        g_ipv4 = self.anm['ipv4']
        eigrp_node = self.anm['eigrp'].node(node)
        node.eigrp.process_id = eigrp_node.process_id
        node.eigrp.custom_config = eigrp_node.custom_config

        ipv4_networks = set()
        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # don't configure IGP for this interface
            ipv4_int = g_ipv4.interface(interface)
            eigrp_int = g_eigrp.interface(interface)
            if not eigrp_int.is_bound:
                continue  # not an EIGRP interface
            network = ipv4_int.subnet
            if eigrp_int and eigrp_int.is_bound and interface.use_ipv4:
                ipv4_networks.add(network)

        # Loopback zero subnet


        ipv4_networks.add(node.loopback_zero.ipv4_cidr)

        node.eigrp.ipv4_networks = sorted(list(ipv4_networks))



#!/usr/bin/python
# -*- coding: utf-8 -*-

import itertools
import logging

import autonetkit
import autonetkit.log as log
from autonetkit.anm.edge import NmEdge
from autonetkit.anm.graph_data import NmGraphData
from autonetkit.anm.interface import NmPort
from autonetkit.anm.node import NmNode
from autonetkit.exception import OverlayNotFound
# TODO: check if this is still a performance hit
from autonetkit.log import CustomAdapter

from autonetkit.anm.ank_element import AnkElement

class OverlayBase(AnkElement):

    '''Base class for overlays - overlay graphs, subgraphs, projections, etc'''

    def __init__(self, anm, overlay_id):
        """"""

        if overlay_id not in anm.overlay_nx_graphs:
            raise OverlayNotFound(overlay_id)

            # TODO: return False instead?

        self._overlay_id = overlay_id
        self._anm = anm
        #logger = logging.getLogger('ANK')
        #logstring = 'Overlay: %s' % str(overlay_id)
        #logger = CustomAdapter(logger, {'item': logstring})
        #object.__setattr__(self, 'log', logger)
        #self.init_logging("graph")
        logger = log
        object.__setattr__(self, 'log', logger)
        self.init_logging("graph")

    def __repr__(self):
        """

        Example:

        >>> anm = autonetkit.topos.house()
        >>> anm['phy']
        phy

        """

        return self._overlay_id

    def is_multigraph(self):
        """
        Example:

        >>> anm = autonetkit.topos.house()
        >>> anm['phy'].is_multigraph()
        False
        >>> anm = autonetkit.topos.multi_edge()
        >>> anm['phy'].is_multigraph()
        True

        """
        return self._graph.is_multigraph()

    @property
    def data(self):
        """Returns data stored on this overlay graph"""

        return NmGraphData(self._anm, self._overlay_id)

    def __contains__(self, n):
        """
        Example:

        >>> anm = autonetkit.topos.house()
        >>> "r1" in anm['phy']
        True
        >>> "test" in anm['phy']
        False
        """

        try:
            return n.node_id in self._graph
        except AttributeError:

            # try with node_id as a string

            return n in self._graph

    def interface(self, interface):
        """"""

        return NmPort(self._anm, self._overlay_id, interface.node_id,
                      interface.interface_id)

    def edge(self, edge_to_find, dst_to_find=None, key=0):
        '''returns edge in this graph with same src and dst
        and key for parallel edges (default is to return first edge)
        #TODO: explain parameter overloading: strings, edges, nodes...

        Example:

        >>> anm = autonetkit.topos.house()
        >>> g_phy = anm['phy']
        >>> e_r1_r2 = g_phy.edge("r1", "r2")

        Can also find from an edge

        >>> e_r1_r2_input = anm['input'].edge(e_r1_r2)

        And for multi-edge graphs can specify key

        >>> anm = autonetkit.topos.multi_edge()
        >>> e1 = anm['phy'].edge("r1", "r2", 0)
        >>> e2 = anm['phy'].edge("r1", "r2", 1)
        >>> e1 == e2
        False

        >>> autonetkit.update_http(anm)
        >>> eth0_r1 = anm["phy"].node("r1").interface("eth0")
        >>> eth3_r1 = anm["phy"].node("r1").interface("eth3")
        >>> eth0_r2 = anm["phy"].node("r2").interface("eth0")
        >>> anm["phy"].has_edge(eth0_r1, eth0_r2)
        True
        >>> anm["phy"].has_edge(eth3_r1, eth0_r2)
        False

        '''

        # TODO: handle multigraphs
        if isinstance(edge_to_find, NmEdge):
            # TODO: tidy this logic
            edge = edge_to_find  # alias for neater code
            if (edge.is_multigraph() and self.is_multigraph()
                and self._graph.has_edge(edge.src,
                                         edge.dst, key=edge.ekey)):
                return NmEdge(self._anm, self._overlay_id,
                              edge.src, edge.dst, edge.ekey)
            elif (self._graph.has_edge(edge.src, edge.dst)):
                return NmEdge(self._anm, self._overlay_id,
                              edge.src, edge.dst)

        if isinstance(edge_to_find, NmEdge):
            src_id = edge_to_find.src
            dst_id = edge_to_find.dst
            search_key = key

            if self.is_multigraph():
                for (src, dst, rkey) in self._graph.edges(src_id,
                                                          keys=True):
                    if dst == dst_id and rkey == search_key:
                        return NmEdge(self._anm, self._overlay_id, src,
                                      dst, search_key)

            for (src, dst) in self._graph.edges(src_id):
                if dst == dst_id:
                    return NmEdge(self._anm, self._overlay_id, src, dst)

        # from here on look for (src, dst) pairs
        src = edge_to_find
        dst = dst_to_find

        if (isinstance(src, basestring) and isinstance(dst, basestring)):
            src = src.lower()
            dst = dst.lower()
            if self.is_multigraph():
                if self._graph.has_edge(src, dst, key=key):
                    return NmEdge(self._anm, self._overlay_id, src,
                                  dst, key)
            elif self._graph.has_edge(src, dst):
                return NmEdge(self._anm, self._overlay_id, src, dst)

        if isinstance(src, NmNode) and isinstance(dst, NmNode):
            src_id = src.node_id
            dst_id = dst.node_id

            if self.is_multigraph():
                if self._graph.has_edge(src_id, dst_id, key):
                    return NmEdge(self._anm, self._overlay_id, src, dst, key)

            else:
                if self._graph.has_edge(src_id, dst_id):
                    return NmEdge(self._anm, self._overlay_id, src, dst)


        if isinstance(src, NmPort) and isinstance(dst, NmPort):
            # further filter result by ports
            src_id = src.node_id
            dst_id = dst.node_id
            src_int = src.interface_id
            dst_int = dst.interface_id


            # TODO: combine duplicated logic from above
            #TODO: test with directed graph

            if self.is_multigraph():
                # search edges from src to dst
                for src, iter_dst, iter_key in self._graph.edges(src_id, keys=True):
                    if iter_dst != dst_id:
                        continue # to a different node

                    ports = self._graph[src][iter_dst][iter_key]["_ports"]
                    if ports[src_id] == src_int and ports[dst_id] == dst_int:
                        return NmEdge(self._anm, self._overlay_id, src_id, dst_id, iter_key)

            else:
                #TODO: add test case for here
                for src, iter_dst in self._graph.edges(src_id):
                    if iter_dst != dst_id:
                        continue # to a different node

                    ports = self._graph[src][iter_dst]["_ports"]
                    if ports[src_id] == src_int and ports[dst_id] == dst_int:
                        return NmEdge(self._anm, self._overlay_id, src_id, dst_id)






    def __getitem__(self, key):
        """"""

        return self.node(key)

    def node(self, key):
        """Returns node based on name
        This is currently O(N). Could use a lookup table

        Example:

        >>> anm = autonetkit.topos.house()
        >>> g_phy = anm['phy']
        >>> r1 = g_phy.node("r1")

        Can also find across layers
        >>> r1_input = anm['input'].node(r1)

        """

        # TODO: refactor

        try:
            if key.node_id in self._graph:
                return NmNode(self._anm, self._overlay_id, key.node_id)
        except AttributeError:

             # try as string id

            if key in self._graph:
                return NmNode(self._anm, self._overlay_id, key)

            # doesn't have node_id, likely a label string, search based on this
            # label

            for node in self:
                if str(node) == key:
                    return node
            # TODO: change warning to an exception
            log.warning('Unable to find node %s in %s ' % (key, self))
            return None

    def overlay(self, key):
        """Get to other overlay graphs in functions"""

        # TODO: refactor: shouldn't be returning concrete instantiation from
        # abstract parent!

        from autonetkit.anm.graph import NmGraph
        return NmGraph(self._anm, key)

    @property
    def name(self):
        """"""

        return self.__repr__()

    def __nonzero__(self):
        return self.anm.has_overlay(self._overlay_id)

    def node_label(self, node):
        """"""

        return repr(NmNode(self._anm, self._overlay_id, node))

    def has_edge(self, edge_to_find, dst_to_find=None,):
        """Tests if edge in graph

        >>> anm = autonetkit.topos.house()
        >>> g_phy = anm['phy']
        >>> r1 = g_phy.node("r1")
        >>> r2 = g_phy.node("r2")
        >>> r5 = g_phy.node("r5")
        >>> g_phy.has_edge(r1, r2)
        True
        >>> g_phy.has_edge(r1, r5)
        False

        >>> e_r1_r2 = anm['input'].edge(r1, r2)
        >>> g_phy.has_edge(e_r1_r2)
        True
        """

        if dst_to_find is None:
            if self.is_multigraph():
                return self._graph.has_edge(edge_to_find.src,
                    edge_to_find.dst, edge_to_find.ekey)

            return self._graph.has_edge(edge_to_find.src, edge_to_find.dst)

        else:
            return bool(self.edge(edge_to_find, dst_to_find))

    def __iter__(self):
        """"""

        return iter(self.nodes())

    def __len__(self):
        """"""

        return len(self._graph)

    def nodes(self, *args, **kwargs):
        """

        >>> anm = autonetkit.topos.multi_as()
        >>> g_phy = anm["phy"]
        >>> g_phy.nodes()
        [r4, r5, r6, r7, r1, r2, r3, r8, r9, r10]

        >>> g_phy.nodes(asn=1)
        [r4, r5, r1, r2, r3]

        >>> g_phy.nodes(asn=3)
        [r7, r8, r9, r10]

        >>> g_phy.nodes(asn=1, ibgp_role="RR")
        [r4, r5]

        >>> g_phy.nodes(asn=1, ibgp_role="RRC")
        [r1, r2, r3]

        """

        result = list(NmNode(self._anm, self._overlay_id, node)
                      for node in self._graph)

        if len(args) or len(kwargs):
            result = self.filter(result, *args, **kwargs)
        return result

    def routers(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be router

        >>> anm = autonetkit.topos.mixed()
        >>> anm['phy'].routers()
        [r1, r2, r3]

        """

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_router()]

    def switches(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be switch

        >>> anm = autonetkit.topos.mixed()
        >>> anm['phy'].switches()
        [sw1]

        """

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_switch()]

    def servers(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be server

        >>> anm = autonetkit.topos.mixed()
        >>> anm['phy'].servers()
        [s1]

        """

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_server()]

    def l3devices(self, *args, **kwargs):
        """Shortcut for nodes(), tests if device is_l3device

        >>> anm = autonetkit.topos.mixed()
        >>> anm['phy'].l3devices()
        [s1, r1, r2, r3]

        """

        result = self.nodes(*args, **kwargs)
        return [r for r in result if r.is_l3device()]

    def device(self, key):
        """To access programatically"""

        return NmNode(self._anm, self._overlay_id, key)

    def groupby(self, attribute, nodes=None):
        """Returns a dictionary sorted by attribute

        >>> anm = autonetkit.topos.house()
        >>> g_phy = anm['phy']
        >>> g_phy.groupby("asn")
        {1: [r1, r2, r3], 2: [r4, r5]}

        Can also specify a subset to work from

        >>> nodes = [n for n in g_phy if n.degree() > 2]
        >>> g_phy.groupby("asn", nodes=nodes)
        {1: [r2, r3]}

        """

        result = {}

        if not nodes:
            data = self.nodes()
        else:
            data = nodes
        data = sorted(data, key=lambda x: x.get(attribute))
        for (key, grouping) in itertools.groupby(data, key=lambda x:
                                                 x.get(attribute)):
            result[key] = list(grouping)

        return result

    def filter(self, nbunch=None, *args, **kwargs):
        """"""

        if nbunch is None:
            nbunch = self.nodes()

        def filter_func(node):
            """Filter based on args and kwargs"""

            return all(getattr(node, key) for key in args) \
                and all(getattr(node, key) == val for (key, val) in
                        kwargs.items())

        return [n for n in nbunch if filter_func(n)]

    def edges(self, src_nbunch=None, dst_nbunch=None, *args,
              **kwargs):
        """
        >>> anm = autonetkit.topos.house()
        >>> g_phy = anm['phy']
        >>> g_phy.edges()
        [(r4, r5), (r4, r2), (r5, r3), (r1, r2), (r1, r3), (r2, r3)]

        >>> g_phy.edge("r1", "r2").color = "red"
        >>> g_phy.edges(color = "red")
        [(r1, r2)]

        """

# src_nbunch or dst_nbunch may be single node
# TODO: refactor this

        if src_nbunch:
            nbunch_out = []
            try:
                src_nbunch = src_nbunch.node_id
            except AttributeError:
                src_nbunch = (n.node_id for n in src_nbunch)

                # only store the id in overlay

        def filter_func(edge):
            """Filter based on args and kwargs"""

            return all(getattr(edge, key) for key in args) \
                and all(getattr(edge, key) == val for (key, val) in
                        kwargs.items())

        if self.is_multigraph():
            valid_edges = list((src, dst, key) for (src, dst, key) in
                               self._graph.edges(src_nbunch, keys=True))
        else:
            default_key = 0
            valid_edges = list((src, dst, default_key)
                               for (src, dst) in self._graph.edges(src_nbunch))

        if dst_nbunch:
            try:
                dst_nbunch = dst_nbunch.node_id
                dst_nbunch = set([dst_nbunch])
            except AttributeError:
                dst_nbunch = (n.node_id for n in dst_nbunch)
                dst_nbunch = set(dst_nbunch)

            valid_edges = list((src, dst, key) for (src, dst, key) in
                               valid_edges if dst in dst_nbunch)

        if len(args) or len(kwargs):
            all_edges = [NmEdge(self._anm, self._overlay_id, src, dst,
                                key) for (src, dst, key) in valid_edges]
            result = list(edge for edge in all_edges
                          if filter_func(edge))
        else:
            result = list(NmEdge(self._anm, self._overlay_id, src, dst,
                                 key) for (src, dst, key) in valid_edges)

        return list(result)

